import sasoptpy as so
import pandas as pd


def test(cas_conn, sols=False):

    # Upload data to server first
    xy_raw = pd.DataFrame([
        [0.0, 1.0],
        [0.5, 0.9],
        [1.0, 0.7],
        [1.5, 1.5],
        [1.9, 2.0],
        [2.5, 2.4],
        [3.0, 3.2],
        [3.5, 2.0],
        [4.0, 2.7],
        [4.5, 3.5],
        [5.0, 1.0],
        [5.5, 4.0],
        [6.0, 3.6],
        [6.6, 2.7],
        [7.0, 5.7],
        [7.6, 4.6],
        [8.5, 6.0],
        [9.0, 6.8],
        [10.0, 7.3]
        ], columns=['x', 'y'])
    xy_data = cas_conn.upload_frame(xy_raw, casout={'name': 'xy_data',
                                                    'replace': True})

    # Read observations
    from sasoptpy.actions import read_data
    POINTS = so.Set(name='POINTS')
    x = so.ParameterGroup(POINTS, name='x')
    y = so.ParameterGroup(POINTS, name='y')
    read_st = read_data(
        table=xy_data,
        index={'target': POINTS, 'key': so.N},
        columns=[
            {'target': x, 'column': 'x'},
            {'target': y, 'column': 'y'}
        ]
    )

    # Parameters and variables
    order = so.Parameter(name='order')
    beta = so.VariableGroup(so.exp_range(0, order), name='beta')
    estimate = so.ImplicitVar(
        (beta[0] + so.expr_sum(beta[k] * x[i] ** k
                                for k in so.exp_range(1, order))
         for i in POINTS), name='estimate')

    surplus = so.VariableGroup(POINTS, name='surplus', lb=0)
    slack = so.VariableGroup(POINTS, name='slack', lb=0)

    objective1 = so.Expression(
        so.expr_sum(surplus[i] + slack[i] for i in POINTS), name='objective1')
    abs_dev_con = so.ConstraintGroup(
        (estimate[i] - surplus[i] + slack[i] == y[i] for i in POINTS),
        name='abs_dev_con')

    minmax = so.Variable(name='minmax')
    objective2 = so.Expression(minmax + 0.0, name='objective2')
    minmax_con = so.ConstraintGroup(
        (minmax >= surplus[i] + slack[i] for i in POINTS), name='minmax_con')

    order.set_init(1)
    L1 = so.Model(name='L1', session=cas_conn)
    L1.set_objective(objective1, sense=so.MIN, name='L1obj')
    L1.include(POINTS, x, y, read_st)
    L1.include(order, beta, estimate, surplus, slack, abs_dev_con)
    L1.add_postsolve_statement('print x y estimate surplus slack;')

    L1.solve(verbose=True)
    if sols:
        sol_data1 = L1.response['Print1.PrintTable'].sort_values('x')
        print(so.get_solution_table(beta))
        print(sol_data1.to_string())

    Linf = so.Model(name='Linf', session=cas_conn)
    Linf.include(L1, minmax, minmax_con)
    Linf.set_objective(objective2, sense=so.MIN, name='Linfobj')

    Linf.solve()
    if sols:
        sol_data2 = Linf.response['Print1.PrintTable'].sort_values('x')
        print(so.get_solution_table(beta))
        print(sol_data2.to_string())

    order.set_init(2)

    L1.solve()
    if sols:
        sol_data3 = L1.response['Print1.PrintTable'].sort_values('x')
        print(so.get_solution_table(beta))
        print(sol_data3.to_string())

    Linf.solve()
    if sols:
        sol_data4 = Linf.response['Print1.PrintTable'].sort_values('x')
        print(so.get_solution_table(beta))
        print(sol_data4.to_string())

    if sols:
        return (sol_data1, sol_data2, sol_data3, sol_data4)
    else:
        return Linf.get_objective_value()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import sasoptpy
from sasoptpy.core import Expression
from sasoptpy.util.package_utils import _to_sas_string


class Parameter(Expression):
    """
    Represents a problem input parameter

    Parameters
    ----------
    name : string
        Name of the parameter
    ptype : string, optional
        Type of the parameter. Possible values are `sasoptpy.STR` and
        `sasoptpy.NUM`
    value : float, optional
        Value of the parameter
    init : float, optional
        Initial value of the parameter

    Examples
    --------

    >>> with so.Workspace('w') as w:
    ...     p = so.Parameter(name='p', init=3)
    ...     p.set_value(5)
    ...
    <sasoptpy.abstract.statement.assignment.Assignment object at 0x7f7952e9bb38>
    >>> print(so.to_optmodel(w))
    proc optmodel;
       num p init 3;
       p = 5;
    quit;

    """

    @sasoptpy.class_containable
    def __init__(self, name, ptype=None, value=None, init=None, **kwargs):
        super().__init__(name=name)
        if name is None:
            name = sasoptpy.util.get_next_name()
        if ptype is None:
            if value is not None and isinstance(value, str):
                ptype = sasoptpy.STR
            elif init is not None and isinstance(init, str):
                ptype = sasoptpy.STR
            else:
                ptype = sasoptpy.NUM
        self._type = ptype
        self._fix_value = value
        self._init = init
        self._parent = None
        self._initialize_self_coef()
        self._abstract = True

    def set_parent(self, parent, key):
        self._parent = parent
        self._key = key

    def _initialize_self_coef(self):
        self.set_member(key=self._name, ref=self, val=1)

    def set_init(self, value):
        self._init = value

    @sasoptpy.containable
    def set_value(self, value):
        self._fix_value = value

    def get_value(self):
        return self._fix_value

    def _expr(self):
        if self._parent:
            return self._parent.get_element_name(self._key)
        return self.get_name()

    def _defn(self):
        if self._parent:
            return None
        else:
            s = '{} {}'.format(self._type, self.get_name())
            if self._init:
                #s += ' init {}'.format(_to_python_string(self._init))
                s += ' init {}'.format(_to_sas_string(self._init))
            elif self._fix_value is not None:
                #s += ' = {}'.format(_to_python_string(self._fix_value))
                s += ' = {}'.format(_to_sas_string(self._fix_value))
            s += ';'
            return s

    def __str__(self):
        return self._name


class ParameterValue(Expression):
    """
    Represents a single value of a parameter

    Parameters
    ----------
    param : Parameter
        Parameter that the value belongs to
    key : tuple, optional
        Key of the parameter value in the multi-index parameter
    prefix : string
        Prefix of the parameter
    suffix : string
        Suffix of the parameter, such as ``.lb`` and ``.ub``

    Notes
    -----

    - Parameter values are mainly used in abstract expressions
    """

    def __init__(self, param, key=None):
        super().__init__()
        self._param = param
        tkey = sasoptpy.util.pack_to_tuple(key)
        self._key = tkey
        self._abstract = True
        self.set_member(key=str(self), ref=self, val=1)

    def __str__(self):
        return \
            sasoptpy.util.package_utils._insert_brackets(
                self._param.get_name(), self._key)

    def _expr(self):
        return str(self)




from collections import OrderedDict
from itertools import product
from math import inf
import warnings
from multiprocessing import Pool

import sasoptpy
from sasoptpy.libs import (pd, np)
from sasoptpy.core import (Expression, Variable, Group)


class VariableGroup(Group):
    """
    Creates a group of :class:`Variable` objects

    Parameters
    ----------
    argv : list, dict, int, :class:`pandas.Index`
        Loop index for variable group
    name : string, optional
        Name (prefix) of the variables
    vartype : string, optional
        Type of variables, `BIN`, `INT`, or `CONT`
    lb : list, dict, :class:`pandas.Series`, optional
        Lower bounds of variables
    ub : list, dict, :class:`pandas.Series`, optional
        Upper bounds of variables
    init : float, optional
        Initial values of variables

    Examples
    --------

    >>> PERIODS = ['Period1', 'Period2', 'Period3']
    >>> production = so.VariableGroup(PERIODS, vartype=so.INT,
                                      name='production', lb=10)
    >>> print(production)
    Variable Group (production) [
      [Period1: production['Period1']]
      [Period2: production['Period2']]
      [Period3: production['Period3']]
    ]

    >>> x = so.VariableGroup(4, vartype=so.BIN, name='x')
    >>> print(x)
    Variable Group (x) [
      [0: x[0]]
      [1: x[1]]
      [2: x[2]]
      [3: x[3]]
    ]

    >>> z = so.VariableGroup(2, ['a', 'b', 'c'], name='z')
    >>> print(z)
    Variable Group (z) [
      [(0, 'a'): z[0, 'a']]
      [(0, 'b'): z[0, 'b']]
      [(0, 'c'): z[0, 'c']]
      [(1, 'a'): z[1, 'a']]
      [(1, 'b'): z[1, 'b']]
      [(1, 'c'): z[1, 'c']]
    ]
    >>> print(repr(z))
    sasoptpy.VariableGroup([0, 1], ['a', 'b', 'c'], name='z')

    Notes
    -----
    * When working with a single model, use the
      :func:`sasoptpy.Model.add_variables` method.
    * If a variable group object is created, it can be added to a model using
      the :func:`sasoptpy.Model.include` method.
    * An individual variable inside the group can be accessed using indices.

      >>> z = so.VariableGroup(2, ['a', 'b', 'c'], name='z', lb=0, ub=10)
      >>> print(repr(z[0, 'a']))
      sasoptpy.Variable(name='z_0_a', lb=0, ub=10, vartype='CONT')

    See also
    --------
    :func:`sasoptpy.Model.add_variables`
    :func:`sasoptpy.Model.include`

    """

    @sasoptpy.class_containable
    def __init__(self, *argv, name, vartype=None, lb=None,
                 ub=None, init=None):
        if len(argv) == 0:
            raise ValueError('An iterable object or None should be given as '
                             'the first parameter')
        super().__init__(name=name)
        self._vardict = OrderedDict()
        self._shadows = OrderedDict()
        self._keyset = []
        self._abstract = False
        self._lb = None
        self._ub = None

        if vartype is None:
            vartype = sasoptpy.CONT

        self._init = init
        self._type = vartype

        if name is None:
            name = sasoptpy.util.get_next_name()

        self._recursive_add_vars(*argv, name=name,
                                 vartype=vartype, lb=lb, ub=ub, init=init,
                                 vardict=self._vardict)
        self.filter_unique_keys()

        lb, ub = sasoptpy.core.util.get_default_bounds_if_none(vartype, lb, ub)

        self.set_bounds(lb=lb, ub=ub, members=False)
        self._objorder = sasoptpy.util.get_creation_id()

        for arg in argv:

            if isinstance(arg, int):
                self._keyset.append(
                    sasoptpy.util.package_utils._extract_argument_as_list(arg))
            else:
                self._keyset.append(
                    sasoptpy.util.package_utils._extract_argument_as_list(arg))
                if not self._abstract and sasoptpy.util.is_set_abstract(arg):
                    self._abstract = True
                    for _, v in self._vardict.items():
                        v._abstract = True

        self._set_var_info()

    def _process_single_var(self, varkey):

        # for varkey in allcombs:
        current_keys = tuple(k for k in varkey)
        is_shadow = any(sasoptpy.abstract.util.is_abstract(i) for i in varkey)
        varname = sasoptpy.core.util.get_name_from_keys(self.get_name(),
                                                        current_keys)
        self._register_keys(current_keys)

        varlb = sasoptpy.util.extract_list_value(current_keys, self._lb)
        varub = sasoptpy.util.extract_list_value(current_keys, self._ub)
        varin = sasoptpy.util.extract_list_value(current_keys, self._init)

        self.add_member(key=current_keys, name=varname, vartype=self._type,
                        lb=varlb, ub=varub, init=varin, shadow=is_shadow)


    def get_name(self):
        """
        Returns the name of the variable group

        Returns
        -------
        name : string
            Name of the variable group

        Examples
        --------

        >>> m = so.Model(name='m')
        >>> var1 = m.add_variables(4, name='x')
        >>> print(var1.get_name())
        x
        """
        return self._name

    def add_member(self, key, name=None, vartype=None, lb=None,
                   ub=None, init=None, shadow=False):
        """
        (Experimental) Adds a new member to Variable Group
        Notes
        -----
        - This method is mainly intended for internal use.
        """

        key = sasoptpy.util.pack_to_tuple(key)

        if lb is None:
            lb = sasoptpy.util.extract_list_value(key, self._lb)
        if ub is None:
            ub = sasoptpy.util.extract_list_value(key, self._ub)
        if init is None:
            init = sasoptpy.util.extract_list_value(key, self._init)
        if vartype is None:
            vartype = self._type

        if shadow is True:
            variable_class = sasoptpy.abstract.ShadowVariable
            dict_to_add = self._shadows
        else:
            variable_class = Variable
            dict_to_add = self._vardict

        if name is None:
            name = sasoptpy.core.util.get_name_from_keys(self.get_name(), key)
        new_var = variable_class(name=name, lb=lb, ub=ub, init=init,
                                 vartype=vartype)
        dict_to_add[key] = new_var
        new_var.set_parent(self)
        return new_var

    def include_member(self, key, var):
        if sasoptpy.core.util.is_variable(var):
            key = sasoptpy.util.pack_to_tuple(key)
            self._vardict[key] = var

    def set_abstract(self, abstract=True):
        self._abstract = abstract

    def _recursive_add_vars(self, *argv, name, vartype, lb, ub, init,
                            vardict, vkeys=(), shadow=False):

        from sasoptpy.util.package_utils import _extract_argument_as_list

        next_arg = _extract_argument_as_list(argv[0])

        for _, i in enumerate(next_arg):
            if isinstance(i, tuple):
                current_keys = vkeys + i
            else:
                current_keys = vkeys + (i,)

            if sasoptpy.abstract.util.is_abstract(i):
                shadow = True

            if len(argv) == 1:
                varname = sasoptpy.core.util.get_name_from_keys(
                    name, current_keys)

                self._register_keys(current_keys)

                varlb = sasoptpy.util.extract_list_value(current_keys, lb)
                varub = sasoptpy.util.extract_list_value(current_keys, ub)
                varin = sasoptpy.util.extract_list_value(current_keys, init)

                self.add_member(key=current_keys, name=varname, vartype=vartype,
                                lb=varlb, ub=varub, init=varin, shadow=shadow)

            else:
                self._recursive_add_vars(*argv[1:], vardict=vardict,
                                         vkeys=current_keys,
                                         name=name, vartype=vartype,
                                         lb=lb, ub=ub, init=init,
                                         shadow=shadow)

    def _set_var_info(self):
        for i in self._vardict:
            self._vardict[i]._set_info(parent=self, key=i)

    def __getitem__(self, key):
        """
        Overloaded method to access individual variables

        Parameters
        ----------
        key : tuple, string or int
            Key of the variable

        Returns
        -------
        ref : :class:`Variable` or list
            Reference to a single Variable or a list of Variable objects

        """
        if self._abstract or sasoptpy.util.is_key_abstract(key) or sasoptpy.core.util.is_expression(key):
            tuple_key = sasoptpy.util.pack_to_tuple(key)
            tuple_key = tuple(i for i in sasoptpy.util.flatten_tuple(tuple_key))
            if tuple_key in self._vardict:
                return self._vardict[tuple_key]
            elif tuple_key in self._shadows:
                return self._shadows[tuple_key]
            else:
                shadow = sasoptpy.abstract.ShadowVariable(self.get_name())
                shadow.set_group_key(self, tuple_key)
                self._shadows[tuple_key] = shadow
                return shadow

        k = sasoptpy.util.pack_to_tuple(key)
        if k in self._vardict:
            return self._vardict[k]
        else:
            indices_to_filter = []
            filter_values = {}
            list_of_variables = []
            for i, _ in enumerate(k):
                if k[i] != '*':
                    indices_to_filter.append(i)
                    filter_values[i] = sasoptpy.util.pack_to_list(k[i])
            for v in self._vardict:
                eligible = True
                for f in indices_to_filter:
                    if v[f] not in filter_values[f]:
                        eligible = False
                if eligible:
                    list_of_variables.append(self._vardict[v])
            if not list_of_variables:
                warnings.warn('Requested variable group is empty:' +
                              ' {}[{}] ({})'.
                              format(self.get_name(), key, type(key)),
                              RuntimeWarning, stacklevel=2)

            if len(list_of_variables) == 0:
                return None
            return list_of_variables

    def __setitem__(self, key, value):
        v = self[key]
        v.set_value(value)
        sasoptpy.abstract.Assignment(v, value)

    def __iter__(self):
        """
        Yields an iterable list of variables inside the variable group

        Returns
        -------
        i : list
            Iterable list of Variable objects
        """
        for v in self._vardict.values():
            yield v

    def _defn(self):
        """
        Returns string to be used in OPTMODEL definition
        """
        name = self.get_name()
        s = 'var {}'.format(name)
        s += ' {'
        for i in self._keyset:
            ind_list = []
            for j in i:
                ind_list.append(
                    sasoptpy.util.package_utils._to_optmodel_quoted_string(j))
            s += '{{{}}}, '.format(','.join(ind_list))
        s = s[:-2]
        s += '} '

        # Grab features
        CONT = sasoptpy.CONT
        BIN = sasoptpy.BIN
        INT = sasoptpy.INT
        if self._type != CONT:
            if self._type == BIN:
                s += 'binary '
            if self._type == INT:
                s += 'integer '

        if sasoptpy.core.util.is_valid_lb(self._lb, self._type):
            s += '>= {} '.format(self._lb)
        if sasoptpy.core.util.is_valid_ub(self._ub, self._type):
            s += '<= {} '.format(self._ub)
        if sasoptpy.core.util.is_valid_init(self._init, self._type):
            s += 'init {} '.format(self._init)

        s = s.rstrip()
        s += ';'

        return(s)

    def _expr(self):
        return self.get_name()

    def _member_defn(self):
        dependents = []
        for v in self.get_members().values():
            dependents.extend(self.get_different_attributes(v))
        defn = sasoptpy.util.get_attribute_definitions(dependents)
        return defn

    def get_members(self):
        """
        Returns a dictionary of members
        """
        return self._vardict

    def get_shadow_members(self):
        return self._shadows

    def get_attributes(self):
        """
        Returns an ordered dictionary of main attributes

        Returns
        --------
        attributes : OrderedDict
            The dictionary consists of `init`, `lb`, and `ub` attributes

        """
        attributes = OrderedDict()
        attributes['init'] = self._init
        attributes['lb'] = self._lb
        attributes['ub'] = self._ub
        return attributes

    def get_type(self):
        """
        Returns the type of variable

        Possible values are:

        * sasoptpy.CONT
        * sasoptpy.INT
        * sasoptpy.BIN

        Examples
        --------

        >>> z = so.VariableGroup(3, name='z', vartype=so.INT)
        >>> z.get_type()
        'INT'

        """
        return self._type

    def get_different_attributes(self, var):
        var_attr = var.get_attributes()
        group_attr = self.get_attributes()

        different_attrs = []
        for key, var_value in var_attr.items():
            if var_value is not None:
                group_value = group_attr.get(key, None)

                def is_equal_to_default(v, k):
                    return v == sasoptpy.core.util.get_default_value(
                        self.get_type(), k)

                if group_value is None and is_equal_to_default(var_value, key):
                    continue

                if sasoptpy.util.is_comparable(group_value) and var_value != group_value:
                    different_attrs.append(
                        {'ref': var, 'key': key, 'value': var_value})
                elif not sasoptpy.util.is_comparable(group_value):
                    different_attrs.append(
                        {'ref': var, 'key': key, 'value': var_value})

        return different_attrs

    def sum(self, *argv):
        """
        Quick sum method for the variable groups

        Parameters
        ----------
        argv : Arguments
            List of indices for the sum

        Returns
        -------
        r : :class:`Expression`
            Expression that represents the sum of all variables in the group

        Examples
        --------

        >>> z = so.VariableGroup(2, ['a', 'b', 'c'], name='z', lb=0, ub=10)
        >>> e1 = z.sum('*', '*')
        >>> print(e1)
        z[1, 'c']  +  z[1, 'a']  +  z[1, 'b']  +  z[0, 'a']  +  z[0, 'b']  +
        z[0, 'c']
        >>> e2 = z.sum('*', 'a')
        >>> print(e2)
         z[1, 'a']  +  z[0, 'a']
        >>> e3 = z.sum('*', ['a', 'b'])
        >>> print(e3)
         z[1, 'a']  +  z[0, 'b']  +  z[1, 'b']  +  z[0, 'a']

        """
        if self._abstract:
            r = Expression()
            symbolic_sum = False
            ind_set = list()
            iter_key = list()
            for i, a in enumerate(argv):
                if isinstance(a, str) and a == '*':
                    sub_list = list()
                    for j in self._keyset[i]:
                        if sasoptpy.abstract.util.is_abstract_set(j):
                            si = sasoptpy.abstract.SetIterator(j)
                            iter_key.append(si)
                            ind_set.append(si)
                            symbolic_sum = True
                        else:
                            #ind_set.append(j)
                            sub_list.append(j)
                    if sub_list:
                        ind_set.append(sub_list)
                else:
                    if hasattr(a, '__iter__'):
                        ind_set.append(a)
                    else:
                        ind_set.append([a])
            combs = product(*ind_set)
            for i in combs:
                var_key = sasoptpy.util.pack_to_tuple(i)
                r = r.add(self[var_key], 1)
            if symbolic_sum:
                #r = r.add(self[tuple(ind_set)])
                r._operator = 'sum'
                r._iterkey = iter_key
            return r
        else:
            r = Expression()
            r.set_temporary()
            feas_set = []
            for i, a in enumerate(argv):
                if a == '*':
                    feas_set.append(self._groups[i])
                elif hasattr(a, "__iter__") and not isinstance(a, str):
                    feas_set.append(a)
                else:
                    feas_set.append([a])
            combs = product(*feas_set)
            for i in combs:
                var_key = sasoptpy.util.pack_to_tuple(i)
                if var_key in self._vardict:
                    r.add(self._vardict[var_key], 1)
            r.set_permanent()
            return r

    def mult(self, vector):
        """
        Quick multiplication method for the variable groups

        Parameters
        ----------
        vector : list, dictionary, :class:`pandas.Series`,\
                 or :class:`pandas.DataFrame`
            Vector to be multiplied with the variable group

        Returns
        -------
        r : :class:`Expression`
            An expression that is the product of the variable group with the
            specified vector

        Examples
        --------

        Multiplying with a list

        >>> x = so.VariableGroup(4, vartype=so.BIN, name='x')
        >>> e1 = x.mult([1, 5, 6, 10])
        >>> print(e1)
         10.0 * x[3]  +  6.0 * x[2]  +  x[0]  +  5.0 * x[1]

        Multiplying with a dictionary

        >>> y = so.VariableGroup([0, 1], ['a', 'b'], name='y', lb=0, ub=10)
        >>> dvals = {(0, 'a'): 1, (0, 'b'): 2, (1, 'a'): -1, (1, 'b'): 5}
        >>> e2 = y.mult(dvals)
        >>> print(e2)
         2.0 * y[0, 'b']  -  y[1, 'a']  +  y[0, 'a']  +  5.0 * y[1, 'b']

        Multiplying with a pandas.Series object

        >>> u = so.VariableGroup(['a', 'b', 'c', 'd'], name='u')
        >>> ps = pd.Series([0.1, 1.5, -0.2, 0.3], index=['a', 'b', 'c', 'd'])
        >>> e3 = u.mult(ps)
        >>> print(e3)
         1.5 * u['b']  +  0.1 * u['a']  -  0.2 * u['c']  +  0.3 * u['d']

        Multiplying with a pandas.DataFrame object

        >>> data = np.random.rand(3, 3)
        >>> df = pd.DataFrame(data, columns=['a', 'b', 'c'])
        >>> print(df)
        NOTE: Initialized model model1
                  a         b         c
        0  0.966524  0.237081  0.944630
        1  0.821356  0.074753  0.345596
        2  0.065229  0.037212  0.136644
        >>> y = m.add_variables(3, ['a', 'b', 'c'], name='y')
        >>> e = y.mult(df)
        >>> print(e)
        0.9665237354418064 * y[0, 'a']  +  0.23708064143289442 * y[0, 'b']  +
        0.944629500537536 * y[0, 'c']  +  0.8213562592159828 * y[1, 'a']  +
        0.07475256894157478 * y[1, 'b']  +  0.3455957019116668 * y[1, 'c']  +
        0.06522945752546017 * y[2, 'a']  +  0.03721153533250843 * y[2, 'b']  +
        0.13664422498043194 * y[2, 'c']

        """

        r = Expression()
        if isinstance(vector, list) or isinstance(vector, np.ndarray):
            for i, key in enumerate(vector):
                var = self._vardict[i, ]
                r._linCoef[var.get_name()] = {'ref': var, 'val': vector[i]}
        elif isinstance(vector, pd.Series):
            for key in vector.index:
                k = sasoptpy.util.pack_to_tuple(key)
                var = self._vardict[k]
                r._linCoef[var.get_name()] = {'ref': var, 'val': vector[key]}
        elif isinstance(vector, pd.DataFrame):
            vectorflat = sasoptpy.util.flatten_frame(vector)
            for key in vectorflat.index:
                k = sasoptpy.util.pack_to_tuple(key)
                var = self._vardict[k]
                r._linCoef[var.get_name()] = {'ref': var, 'val': vectorflat[key]}
        else:
            for i, key in enumerate(vector):
                if isinstance(key, tuple):
                    k = key
                else:
                    k = (key,)
                var = self._vardict[k]
                try:
                    r._linCoef[var.get_name()] = {'ref': var, 'val': vector[i]}
                except KeyError:
                    r._linCoef[var.get_name()] = {'ref': var, 'val': vector[key]}

        return r

    def set_init(self, init):
        """
        Specifies or updates the initial values

        Parameters
        ----------
        init : float, list, dict, :class:`pandas.Series`
            Initial value of the variables

        Examples
        --------

        >>> m = so.Model(name='m')
        >>> y = m.add_variables(3, name='y')
        >>> print(y._defn())
        var y {{0,1,2}};
        >>> y.set_init(5)
        >>> print(y._defn())
        var y {{0,1,2}} init 5;

        """
        self._init = init
        for v in self._vardict:
            inval = sasoptpy.util.extract_list_value(v, init)
            self._vardict[v].set_init(inval)
        for v in self._shadows:
            self._shadows[v].set_init(init)

    def set_bounds(self, lb=None, ub=None, members=True):
        """
        Specifies or updates bounds for the variable group

        Parameters
        ----------
        lb : float, :class:`pandas.Series`, optional
            Lower bound
        ub : float, :class:`pandas.Series`, optional
            Upper bound

        Examples
        --------

        >>> z = so.VariableGroup(2, ['a', 'b', 'c'], name='z', lb=0, ub=10)
        >>> print(repr(z[0, 'a']))
        sasoptpy.Variable(name='z_0_a', lb=0, ub=10, vartype='CONT')
        >>> z.set_bounds(lb=3, ub=5)
        >>> print(repr(z[0, 'a']))
        sasoptpy.Variable(name='z_0_a', lb=3, ub=5, vartype='CONT')

        >>> u = so.VariableGroup(['a', 'b', 'c', 'd'], name='u')
        >>> lb_vals = pd.Series([1, 4, 0, -1], index=['a', 'b', 'c', 'd'])
        >>> u.set_bounds(lb=lb_vals)
        >>> print(repr(u['b']))
        sasoptpy.Variable(name='u_b', lb=4, ub=inf, vartype='CONT')

        """

        if lb is not None:
            self._lb = sasoptpy.core.util.get_group_bound(lb)
        if ub is not None:
            self._ub = sasoptpy.core.util.get_group_bound(ub)

        if members:
            for v in self._vardict:
                varlb = sasoptpy.util.extract_list_value(v, lb)
                if lb is not None:
                    self[v].set_bounds(lb=varlb)
                varub = sasoptpy.util.extract_list_value(v, ub)
                if ub is not None:
                    self[v].set_bounds(ub=varub)

    def set_member_value(self, key, value):
        pass

    def get_member_by_name(self, name):
        keys = sasoptpy.abstract.util.get_key_from_name(name)
        return self[keys]

    def __str__(self):
        """
        Generates a representation string
        """
        s = 'Variable Group ({}) [\n'.format(self.get_name())
        for k in self._vardict:
            v = self._vardict[k]
            s += '  [{}: {}]\n'.format(sasoptpy.util.get_first_member(k), v)
        s += ']'
        return s

    def __repr__(self):
        """
        Returns a string representation of the object.
        """
        s = 'sasoptpy.VariableGroup('
        keylen = max(map(len, self._vardict))
        for i in range(keylen):
            ls = []
            for k in self._vardict:
                if k[i] not in ls:
                    ls.append(k[i])
            s += '{}, '.format(ls)
        s += 'name=\'{}\')'.format(self.get_name())
        return s



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for bug-fixes and patch releases.
"""

import unittest
import sasoptpy as so
from inspect import cleandoc


class TestPatches(unittest.TestCase):
    """
    Unit tests for patches
    """

    def setUp(self):
        so.reset()

    def test_multi_iterators(self):

        def get_nearby(v):
            return (v-1, v+1)

        G = list(range(5))

        m = so.Model(name='test_multi_iter')
        x = m.add_variables(G, name='x', ub=10)
        m.set_objective(so.expr_sum(x[i] for i in G), name='obj',
                        sense=so.maximize)

        m.add_constraints(
            (so.expr_sum(x[k] for k in range(j) if k != i) >= x[i] -1
             for i in G for j in get_nearby(i) if j >=0 and j<=4),
            name='c1'
        )

        for i in G:
            for j in get_nearby(i):
                if j>=0 and j<=4:
                    m.add_constraint(so.expr_sum(x[k] for k in range(j) if k!=i) >= x[i]-1, name=so.util.get_next_name())

        m.add_constraints(
            (so.expr_sum(x[k] for k in range(j)) >= x[i] - 1
             for i in range(5) for j in get_nearby(i) if j >= 0 and j <= 4),
            name='c2'
        )

        self.assertEqual(so.to_optmodel(m), cleandoc('''
            proc optmodel;
               var x {{0,1,2,3,4}} <= 10;
               max obj = x[0] + x[1] + x[2] + x[3] + x[4];
               con c1_0_1 : - x[0] >= -1;
               con c1_1_0 : - x[1] >= -1;
               con c1_1_2 : x[0] - x[1] >= -1;
               con c1_2_1 : x[0] - x[2] >= -1;
               con c1_2_3 : x[0] + x[1] - x[2] >= -1;
               con c1_3_2 : x[0] + x[1] - x[3] >= -1;
               con c1_3_4 : x[0] + x[1] + x[2] - x[3] >= -1;
               con c1_4_3 : x[0] + x[1] + x[2] - x[4] >= -1;
               con o37 : - x[0] >= -1;
               con o40 : - x[1] >= -1;
               con o43 : x[0] - x[1] >= -1;
               con o46 : x[0] - x[2] >= -1;
               con o49 : x[0] + x[1] - x[2] >= -1;
               con o52 : x[0] + x[1] - x[3] >= -1;
               con o55 : x[0] + x[1] + x[2] - x[3] >= -1;
               con o58 : x[0] + x[1] + x[2] - x[4] >= -1;
               con c2_0_1 : 1 >= -1;
               con c2_1_0 : - x[1] >= -1;
               con c2_1_2 : x[0] >= -1;
               con c2_2_1 : x[0] - x[2] >= -1;
               con c2_2_3 : x[0] + x[1] >= -1;
               con c2_3_2 : x[0] + x[1] - x[3] >= -1;
               con c2_3_4 : x[0] + x[1] + x[2] >= -1;
               con c2_4_3 : x[0] + x[1] + x[2] - x[4] >= -1;
               solve;
            quit;'''))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#


from contextlib import contextmanager
from functools import wraps, partial
import warnings

import sasoptpy


def containable(func=None, standalone=True):
    if func is None:
        return partial(containable, standalone=standalone)

    @wraps(func)
    def wrapper(*args, **kwargs):

        if not standalone and not sasoptpy.container:
            warnings.warn('This function is not intended to be used without any'
                          ' container', UserWarning)

        if sasoptpy.container:
            try:
                statement_func = sasoptpy.statement_dictionary[wrapper]
            except KeyError:
                raise NotImplementedError(
                    'Container support for {} is not implemented'.format(
                        func.__name__))
            statements = statement_func(*args, **kwargs)
            if isinstance(statements, list):
                for st in statements:
                    append_to_container(st)
            else:
                append_to_container(statements)
            return statements
        else:
            return func(*args, **kwargs)
    return wrapper


def class_containable(func):
    def class_append(*args, **kwargs):
        is_internal = kwargs.pop('internal', None)
        func(*args, **kwargs)
        if sasoptpy.container and is_internal is None:
            sasoptpy.container.append(args[0])

    return class_append


def append_to_container(statement):
    if hasattr(statement, 'is_internal') and statement.is_internal():
        pass
    else:
        sasoptpy.container.append(statement)


@contextmanager
def under_condition(c):
    if type(c) == bool:
        yield c
        return True
    original = sasoptpy.conditions
    if original is None:
        sasoptpy.conditions = []
    sasoptpy.conditions = sasoptpy.conditions + [c]
    yield
    sasoptpy.conditions = original


def inline_condition(c):
    if sasoptpy.container_conditions:
        sasoptpy.container.sym.add_condition(c)
        return True
    return False


@contextmanager
def set_container(s, conditions=False):
    original = sasoptpy.container
    sasoptpy.container = s
    cond_original = None
    if conditions:
        cond_original = sasoptpy.container_conditions
        sasoptpy.container_conditions = True

    yield

    sasoptpy.container = original
    if conditions:
        sasoptpy.container_conditions = cond_original





from collections import OrderedDict
import sasoptpy
from .condition import Conditional, Condition

class SetIterator(sasoptpy.Expression):
    """
    Creates an iterator object for a given Set

    Parameters
    ----------
    initset : :class:`Set`
        Set to be iterated on
    name : string, optional
        Name of the iterator
    datatype : string, optional
        Type of the iterator

    Notes
    -----

    - :class:`abstract.SetIterator` objects are created automatically when
      iterating over a :class:`abstract.Set` object

    Examples
    --------

    >>> S = so.Set(name='S')
    >>> for i in S:
    ...     print(i.get_name(), type(i))
    o19 <class 'sasoptpy.abstract.set_iterator.SetIterator'>

    """

    def __init__(self, initset, name=None, datatype=None):
        if name is None:
            name = sasoptpy.util.get_next_name()
        super().__init__(name=name)
        self.set_member(key=name, ref=self, val=1.0)
        self._set = initset
        if datatype is None:
            datatype = sasoptpy.NUM
        self._type = datatype
        self.sym = Conditional(self)

    def get_set(self):
        return self._set

    def get_type(self):
        return self._type

    def __hash__(self):
        return hash('{}'.format(id(self)))

    def _get_for_expr(self):
        return '{} in {}'.format(self._expr(),
                                 sasoptpy.to_expression(self._set))

    def _expr(self):
        return self.get_name()

    def __str__(self):
        return self._name

    def _defn(self):
        return self._get_for_expr()

    def __repr__(self):
        s = 'sasoptpy.SetIterator({}, name=\'{}\')'.format(self._set, self.get_name())
        return s

    # def _cond_expr(self):
    #     return '<' + self._expr() + '>'

    def __lt__(self, other):
        return Condition(self, '<', other)

    def __gt__(self, other):
        return Condition(self, '>', other)

    def __le__(self, other):
        return Condition(self, '<=', other)

    def __ge__(self, other):
        return Condition(self, '>=', other)

    def __eq__(self, other):
        return Condition(self, 'EQ', other)

    def __ne__(self, other):
        return Condition(self, 'NE', other)


class SetIteratorGroup(OrderedDict, sasoptpy.Expression):
    """
    Creates a group of set iterator objects for multi-dimensional sets

    Parameters
    ----------
    initset : :class:`Set`
        Set to be iterated on
    names : string, optional
        Names of the iterators
    datatype : string, optional
        Types of the iterators

    Examples
    --------

    >>> T = so.Set(name='T', settype=[so.STR, so.NUM])
    >>> for j in T:
    ...     print(j.get_name(), type(j))
    ...     for k in j:
    ...         print(k.get_name(), type(k))
    o5 <class 'sasoptpy.abstract.set_iterator.SetIteratorGroup'>
    o6 <class 'sasoptpy.abstract.set_iterator.SetIterator'>
    o8 <class 'sasoptpy.abstract.set_iterator.SetIterator'>

    """

    def __init__(self, initset, datatype=None, names=None):
        super(SetIteratorGroup, self).__init__()
        self._objorder = sasoptpy.util.get_creation_id()
        self._name = sasoptpy.util.get_next_name()
        self._set = initset
        self._init_members(names, datatype)
        self.sym = sasoptpy.abstract.Conditional(self)

    def _init_members(self, names, datatype):
        if names is not None:
            for i, name in enumerate(names):
                dt = datatype[i] if datatype is not None else None
                it = SetIterator(None, name=name, datatype=dt)
                self.append(it)
        else:
            for i in datatype:
                it = SetIterator(None, datatype=i)
                self.append(it)

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            if isinstance(key, int):
                return list(self.values())[key]

    def get_name(self):
        return self._name

    def append(self, object):
        name = object.get_name()
        self[name] = object

    def _get_for_expr(self):
        #return '<{}> in {}'.format(self._expr(), self._set._name)
        comb = '<' + ', '.join(str(i) for i in self.values()) + '>'
        s = '{} in {}'.format(comb, sasoptpy.to_expression(self._set))
        return s

    def _expr(self):
        return ', '.join(str(i) for i in self.values())

    def _defn(self):
        return self._get_for_expr()


    def __iter__(self):
        for i in self.values():
            yield i

    def __repr__(self):
        return 'sasoptpy.SetIteratorGroup({}, datatype=[{}], names=[{}])'.format(
            self._set,
            ', '.join('\'' + i.get_type() + '\'' for i in self.values()),
            ', '.join('\'' + i.get_name() + '\'' for i in self.values())
        )

    def __str__(self):
        s = ', '.join(str(i) for i in self.values())
        return '(' + s + ')'

    def __hash__(self):
        hashstr = ','.join(str(id(i)) for i in self.values())
        return hash(hashstr)



"""
This file keeps a copy of expected responses to unit tests.
"""

from inspect import cleandoc
import os
current_dir = os.path.dirname(os.path.abspath(__file__))


def read_file(problem):
    file_path = os.path.join(current_dir, 'responses/{}.sas'.format(problem))
    return open(file_path, 'r').read()


# Food Manufacture 1
fm1 = [
    read_file('food_manufacture_1_0'),
    read_file('food_manufacture_1_1'),
    read_file('food_manufacture_1_2'),
    read_file('food_manufacture_1_3')
]

# Food Manufacture 2
fm2 = [
    read_file('food_manufacture_2')
]

# Factory Planning 1
fp1 = [
    read_file('factory_planning_1')
]

# Factory Planning 2
fp2 = [
    read_file('factory_planning_2')
]

# Manpower Planning
mp = [
    read_file('manpower_planning_0'),
    read_file('manpower_planning_1')
]

# Refinery Optimization
ro = [
    read_file('refinery_optimization')
]

# Mining Optimization
mo = [
    read_file('mining_optimization')
]

# Farm Planning
farmp = [
    read_file('farm_planning')
]

# Economic Planning
econ = [
    read_file('economic_planning_0'),
    read_file('economic_planning_1'),
    read_file('economic_planning_2'),
    read_file('economic_planning_3')
]

# Decentralization
dc = [
    read_file('decentralization_0'),
    read_file('decentralization_1')
]

# Optimal Wedding
ow = [
    read_file('sas_optimal_wedding')
]

# Kidney Exchange
kx = [
    read_file('sas_kidney_exchange_0'),
    read_file('sas_kidney_exchange_1')
]

# Multiobjective
multiobj = [
    read_file('multiobjective')
]

# Curve Fitting
cf = [
    read_file('curve_fitting_0'),
    read_file('curve_fitting_1'),
    read_file('curve_fitting_2'),
    read_file('curve_fitting_3')
]

# Nonlinear 1
nl1 = [
    read_file('nonlinear_1')
]

# Nonlinear 2
nl2 = [
    read_file('nonlinear_2_0'),
    read_file('nonlinear_2_1')
]

least_squares = [
    read_file('least_squares')
]

efficiency_analysis = [
    read_file('efficiency_analysis')
]



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for core classes.
"""

import unittest
from inspect import cleandoc

import sasoptpy as so


class TestObjective(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.Objective` objects
    """

    @classmethod
    def setUpClass(cls):
        cls.model = so.Model(name='test_objective')
        cls.x = cls.model.add_variable(name='x')

    def test_init(self):
        x = TestObjective.x
        all_params = so.Objective(x ** 2 + 2 * x - 1, name='all_params',
                                  sense=so.MIN)
        self.assertEqual(str(all_params), "(x) ** (2) + 2 * x - 1")

        no_sense = so.Objective(2 * x, name='no_sense')
        self.assertEqual(no_sense.get_sense(), so.MIN)

    def test_model_init(self):
        m = TestObjective.model
        x = TestObjective.x

        obj = m.set_objective(2 * x - x ** 3, sense=so.MIN, name='new_obj')
        self.assertEqual(str(m.get_objective()), "2 * x - ((x) ** (3))")

    def test_defn(self):
        x = TestObjective.x
        def_obj = so.Objective(4 * x - x, name='objdef', sense=so.MAX)
        self.assertEqual(so.to_definition(def_obj),
                         "max objdef = 3 * x;")

    def test_sense(self):
        m = TestObjective.model
        x = TestObjective.x

        m.set_objective(3 * x, name='obj1')
        self.assertEqual(m.get_objective().get_sense(), so.MIN)
        m.get_objective().set_sense(so.MAX)
        self.assertEqual(m.get_objective().get_sense(), so.MAX)

    def test_in_container(self):
        from sasoptpy.actions import for_loop, solve

        with so.Workspace('w') as w:
            x = so.Variable(name='x', lb=0.5, vartype=so.INT, init=2)
            o = so.Objective(x**2, name='obj', sense=so.MIN)
            m = so.Model(name='model1')
            m.include(x, o)
            m.solve(options={'with': 'milp', 'relaxint': True}, primalin=True)
        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x integer >= 0.5 init 2;
               min obj = (x) ^ (2);
               problem model1 include x obj;
               use problem model1;
               solve with milp relaxint / primalin;
            quit;'''))

    def test_value(self):
        x = TestObjective.x
        all_params = so.Objective(x ** 2 + 2 * x - 1, name='all_params',
                                  sense=so.MIN)
        x.set_value(2)
        self.assertEqual(all_params.get_value(), 7)

        y = so.Variable(name='y')
        y._abstract = True
        new_obj = so.Objective(y ** 2 - 1, name='new_obj')
        self.assertRaises(ValueError, new_obj.get_value)


    @classmethod
    def tearDownClass(self):
        so.reset()





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for workspaces.
"""

import os
import unittest
import warnings
from inspect import cleandoc

from swat import CAS, SWATError
import sasoptpy as so

from tests.swat_config import create_cas_connection


class TestWorkspace(unittest.TestCase):
    """
    Unit tests for the :class:`sasoptpy.Workspace` objects
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def setUp(self):
        so.reset()

    def test_vg_in_session(self):
        with so.Workspace('w') as w:
            x = so.VariableGroup(4, name='x')
            x[2].set_bounds(lb=1)
            x[3].set_init(5)
        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x {{0,1,2,3}};
               x[3] = 5;
               x[2].lb = 1;
            quit;'''))

    def test_attributes(self):
        with so.Workspace('w') as w:
            S = so.Set(name='S')
            x = so.Variable(name='x')
            y = so.VariableGroup(4, name='y')
            z = so.VariableGroup(S, name='z')
            d = so.Variable(name='d')
            e = so.Variable(name='d')
            t = so.VariableGroup(2, name='t')
            u = so.VariableGroup(2, name='t')

        self.assertIn('Workspace[ID', str(w))
        self.assertEqual('sasoptpy.Workspace(w)', repr(w))
        w.set_session(None)
        self.assertEqual(w.get_session(), None)
        self.assertIs(x, w.get_variable('x'))
        self.assertIs(y, w.get_variable('y'))
        self.assertIs(z[0], w.get_variable('z[0]'))
        def warn_duplicate():
            w.get_variable('d')
        self.assertWarns(UserWarning, warn_duplicate)
        def warn_duplicate_vg():
            w.get_variable('t')
        self.assertWarns(UserWarning, warn_duplicate_vg)
        w.set_variable_value('z[0]', 1)
        self.assertEqual(z[0].get_value(), 1)
        self.assertEqual(w.get_variable('a'), None)

    def test_var_values(self):
        if TestWorkspace.conn is None:
            self.skipTest('CAS session is not available')

        from sasoptpy.actions import solve

        with so.Workspace('test_var_vals', session=TestWorkspace.conn) as w:
            S = so.Set(name='S', value=[1, 2, 3])
            x = so.Variable(name='x', lb=1, ub=4)
            y = so.VariableGroup(S, name='y', ub=7)
            z = so.VariableGroup(2, name='z', ub=2)
            o = so.Objective(x+y[1]+y[2]+y[3]+z[0], name='obj', sense=so.MAX)
            solve()

        w.submit(verbose=True)
        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S = {1,2,3};
               var x >= 1 <= 4;
               var y {{S}} <= 7;
               var z {{0,1}} <= 2;
               max obj = x + y[1] + y[2] + y[3] + z[0];
               solve;
            quit;'''))
        self.assertEqual(x.get_value(), 4)
        self.assertEqual(y[1].get_value(), 7)
        self.assertEqual(z[0].get_value(), 2)

    def test_ws_parsing(self):
        if TestWorkspace.conn is None:
            self.skipTest('CAS session is not available')

        from sasoptpy.actions import solve, drop, print_item
        from math import inf

        with so.Workspace('test_ws_parsing', session=TestWorkspace.conn)\
                as w:
            x = so.Variable(name='x')
            y = so.Variable(name='y', vartype=so.INT, lb=-inf)
            o = so.Objective(x**2-4*x+4, sense=so.MIN, name='obj')
            c1 = so.Constraint(x <= 1, name='c1')
            c2 = so.Constraint(x == 3*y, name='c2')
            s1 = solve(options={'with': so.BLACKBOX})
            p1 = print_item(x)
            drop(c1)
            s2 = so.LiteralStatement('solve with blackbox;')
            p2 = so.LiteralStatement('print y;')

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x;
               var y integer;
               min obj = (x) ^ (2) - 4 * x + 4;
               con c1 : x <= 1;
               con c2 : x - 3 * y = 0;
               solve with blackbox;
               print x;
               drop c1;
               solve with blackbox;
               print y;
            quit;'''))

        w.submit()
        self.assertEqual(x.get_value(), 3)
        print(s1.get_problem_summary().to_string())
        self.assertEqual(s1.get_problem_summary().to_string(), cleandoc('''
            Problem Summary

                                            Value
            Label                                
            Objective Sense          Minimization
            Objective Function                obj
            Objective Type              Quadratic
                                                 
            Number of Variables                 2
            Bounded Above                       0
            Bounded Below                       0
            Bounded Below and Above             0
            Free                                2
            Fixed                               0
            Binary                              0
            Integer                             1
                                                 
            Number of Constraints               2
            Linear LE (<=)                      1
            Linear EQ (=)                       1
            Linear GE (>=)                      0
            Linear Range                        0
                                                 
            Constraint Coefficients             3'''))
        self.assertEqual(p1.get_response().to_string(), cleandoc('''
            x

                 x
            0  0.0'''))
        self.assertEqual(p2.get_response().to_string(), cleandoc('''
            y
            
                 y
            0  1.0'''))


    def test_multithread_workspace(self):

        import time
        from threading import Thread

        def create_workspace(i):

            with so.Workspace(f'w{i}') as w:
                self.assertEqual(so.container, w)
                print('Start workspace: {}'.format(w.name))
                time.sleep(1)

            print('Exit workspace: {}'.format(w.name))
            return i

        threads = []
        for j in [1, 2, 3]:
            t = Thread(target=create_workspace, args=(j,))
            threads.append(t)
            t.start()

        for thread in threads:
            thread.join()

        print(threads)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for FOR loops.
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../..')))
from util import assert_equal_wo_temps

from sasoptpy.actions import for_loop


class TestForLoop(unittest.TestCase):
    """
    Unit tests for FOR statements
    """

    def setUp(self):
        so.reset()

    def testRegularWorkspace(self):

        from sasoptpy.actions import for_loop, solve, set_value, set_objective, print_item

        with so.Workspace('ex_9_1_matirx_sqrt', session=None) as w:

            so.LiteralStatement('call streaminit(1);')

            n = so.Parameter(name='n', value=5)
            rn = so.Set(name='RN', value=so.exp_range(1, n))
            A = so.ParameterGroup(rn, rn, name='A', value="10-20*rand('UNIFORM')")
            P = so.ParameterGroup(rn, rn, name='P')

            for i in for_loop(rn):
                for j in for_loop(so.exp_range(i, n)):
                    set_value(P[i, j], so.expr_sum(A[i, k] * A[j, k] for k in rn))

            q = so.VariableGroup(rn, rn, name='q')
            set_value(q[1, 1], 1)
            set_objective(
                so.expr_sum((
                    so.expr_sum(q[k, i] * q[k, j] for k in so.exp_range(1, i)) +
                    so.expr_sum(q[i, k] * q[k, j] for k in so.exp_range(i+1, j)) +
                    so.expr_sum(q[i, k] * q[j, k] for k in so.exp_range(j+1, n)) -
                    P[i, j]
                )**2 for i in rn for j in so.exp_range(i, n))
                , name='r', sense=so.MIN)

            solve()
            print_item(P)
            print_item(q)

        assert_equal_wo_temps(
            self, so.to_optmodel(w),
            cleandoc(
                '''
                proc optmodel;
                   call streaminit(1);
                   num n = 5;
                   set RN = 1..n;
                   num A {RN, RN} = 10-20*rand('UNIFORM');
                   num P {RN, RN};
                   for {TEMP1 in RN} do;
                      for {TEMP2 in TEMP1..n} do;
                         P[TEMP1, TEMP2] = sum {k in RN} (A[TEMP1, k] * A[TEMP2, k]);
                      end;
                   end;
                   var q {{RN}, {RN}};
                   q[1, 1] = 1;
                   MIN r = sum {i in RN, j in i..n} ((sum {k in 1..i} (q[k, i] * q[k, j]) + sum {k in i+1..j} (q[i, k] * q[k, j]) + sum {k in j+1..n} (q[i, k] * q[j, k]) - P[i, j]) ^ (2));
                   solve;
                   print P;
                   print q;
                quit;
                '''))

    def test_basic(self):

        from sasoptpy.actions import put_item

        with so.Workspace('w') as w:
            for i in for_loop(range(1, 3)):
                for j in for_loop(['a', 'b']):
                    put_item(i, j)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc("""
            proc optmodel;
               for {TEMP1 in 1..2} do;
                  for {TEMP2 in {'a','b'}} do;
                     put TEMP1 TEMP2;
                  end;
               end;
            quit;"""))

    def test_with_assignment(self):
        with so.Workspace('w') as w:
            r = so.exp_range(1, 11)
            x = so.VariableGroup(r, name='x')
            for i in for_loop(r):
                x[i] = 1

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc("""
            proc optmodel;
               var x {{1,2,3,4,5,6,7,8,9,10}};
               for {TEMP1 in 1..10} do;
                  x[TEMP1] = 1;
               end;
            quit;"""))

    def test_with_predefined_set(self):
        with so.Workspace('w') as w:
            cn = so.Set(name='C', value=['a', 'b', 'c'])
            for i in for_loop(cn):
                so.actions.put_item(i)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc("""
            proc optmodel;
               set C = {'a','b','c'};
               for {TEMP1 in C} do;
                  put TEMP1;
               end;
            quit;"""))

    def test_with_create_data(self):
        from sasoptpy.actions import create_data, read_data
        from sasoptpy.util import concat, iterate
        with so.Workspace('w') as w:
            m = so.Set(name='m', value=range(1, 4))
            rev = so.VariableGroup(range(1, 13), name='revenue')
            read_data(
                table='revdata',
                index={'key': so.N},
                columns=[
                    {'column': 'rev', 'target': rev}
                ]
            )
            month = so.SetIterator(None, name='month')
            for q in for_loop(so.exp_range(1, 5)):
                create_data(
                    table=concat('qtr', q),
                    index={'key': [month], 'set': m},
                    columns=[
                        {'expression': rev[month+(q-1)*3]}
                    ]
                )

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc("""
            proc optmodel;
               set m = 1..3;
               var revenue {{1,2,3,4,5,6,7,8,9,10,11,12}};
               read data revdata into [_N_] revenue=rev;
               for {TEMP1 in 1..4} do;
                  create data ('qtr' || TEMP1) from [month] = {m} revenue[month + 3.0 * TEMP1 - 3];
               end;
            quit;"""))


    def test_multiple_set_loop(self):

        with so.Workspace('w') as w:
            r = so.Set(name='R', value=range(1, 11))
            c = so.Set(name='C', value=range(1, 6))
            a = so.ParameterGroup(r, c, name='A', ptype=so.number)
            for (i, j) in for_loop(r, c):
                a[i, j] = 1

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set R = 1..10;
               set C = 1..5;
               num A {R, C};
               for {TEMP1 in R, TEMP2 in C} do;
                  A[TEMP1, TEMP2] = 1;
               end;
            quit;'''))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
SAS Optimization Interface for Python (sasoptpy)
************************************************

This file includes static methods and imports.

"""

from sasoptpy.libs import *
import sasoptpy.util

from sasoptpy.util import (
    quick_sum, expr_sum, expr_min, reset, reset, reset_globals,
    flatten_frame, get_value_table,
    get_solution_table, dict_to_frame, exp_range,
    to_expression, to_definition, to_optmodel, is_linear,
    load_package_globals, export_to_mps)

from sasoptpy.structure import (containable, class_containable, set_container)

from sasoptpy.core.expression import Expression
from sasoptpy.core import (Variable, VariableGroup, Constraint, ConstraintGroup,
                           Model, Objective, Auxiliary, Symbol)

from sasoptpy.abstract import (Set, InlineSet, Parameter, ImplicitVar,
                               ParameterGroup, SetIterator,
                               SetIteratorGroup, LiteralStatement)

load_package_globals()

import sasoptpy.config
from sasoptpy.config import Config, _load_default_config
_load_default_config()

statement_dictionary = dict()
container = None
from sasoptpy.util import load_function_containers
load_function_containers()
conditions = None
from sasoptpy.actions import register_actions
register_actions()


import sasoptpy.interface
sasoptpy.mediators = dict()
from sasoptpy.util import load_default_mediators
load_default_mediators()


from sasoptpy.session import Workspace

name = "sasoptpy"
from sasoptpy.version import __version__



import re

def get_generic_form(string):
    """
    Replaces temporary object names (in the form of o1, o2, ...) for
    comparison purposes
    """
    matches = re.findall(r'o[\d]+', string, re.MULTILINE)
    unique_matches = sorted(set(matches), key=lambda x: int(x[1:]))
    temp_ctr = 0
    for i in unique_matches:
        temp_ctr += 1
        string = string.replace(i, 'TEMP{}'.format(temp_ctr))
    return string


def assert_equal_wo_temps(tester, left, right):
    left_wo_temp = get_generic_form(left)
    right_wo_temp = get_generic_form(right)
    return tester.assertEqual(left_wo_temp, right_wo_temp)



import sasoptpy as so
import pandas as pd
import numpy as np


def test(cas_conn, **kwargs):

    m = so.Model(name='refinery_optimization', session=cas_conn)

    crude_data = pd.DataFrame([
        ['crude1', 20000],
        ['crude2', 30000]
        ], columns=['crude', 'crude_ub']).set_index(['crude'])

    arc_data = pd.DataFrame([
        ['source', 'crude1', 6],
        ['source', 'crude2', 6],
        ['crude1', 'light_naphtha', 0.1],
        ['crude1', 'medium_naphtha', 0.2],
        ['crude1', 'heavy_naphtha', 0.2],
        ['crude1', 'light_oil', 0.12],
        ['crude1', 'heavy_oil', 0.2],
        ['crude1', 'residuum', 0.13],
        ['crude2', 'light_naphtha', 0.15],
        ['crude2', 'medium_naphtha', 0.25],
        ['crude2', 'heavy_naphtha', 0.18],
        ['crude2', 'light_oil', 0.08],
        ['crude2', 'heavy_oil', 0.19],
        ['crude2', 'residuum', 0.12],
        ['light_naphtha', 'regular_petrol', np.nan],
        ['light_naphtha', 'premium_petrol', np.nan],
        ['medium_naphtha', 'regular_petrol', np.nan],
        ['medium_naphtha', 'premium_petrol', np.nan],
        ['heavy_naphtha', 'regular_petrol', np.nan],
        ['heavy_naphtha', 'premium_petrol', np.nan],
        ['light_naphtha', 'reformed_gasoline', 0.6],
        ['medium_naphtha', 'reformed_gasoline', 0.52],
        ['heavy_naphtha', 'reformed_gasoline', 0.45],
        ['light_oil', 'jet_fuel', np.nan],
        ['light_oil', 'fuel_oil', np.nan],
        ['heavy_oil', 'jet_fuel', np.nan],
        ['heavy_oil', 'fuel_oil', np.nan],
        ['light_oil', 'light_oil_cracked', 2],
        ['light_oil_cracked', 'cracked_oil', 0.68],
        ['light_oil_cracked', 'cracked_gasoline', 0.28],
        ['heavy_oil', 'heavy_oil_cracked', 2],
        ['heavy_oil_cracked', 'cracked_oil', 0.75],
        ['heavy_oil_cracked', 'cracked_gasoline', 0.2],
        ['cracked_oil', 'jet_fuel', np.nan],
        ['cracked_oil', 'fuel_oil', np.nan],
        ['reformed_gasoline', 'regular_petrol', np.nan],
        ['reformed_gasoline', 'premium_petrol', np.nan],
        ['cracked_gasoline', 'regular_petrol', np.nan],
        ['cracked_gasoline', 'premium_petrol', np.nan],
        ['residuum', 'lube_oil', 0.5],
        ['residuum', 'jet_fuel', np.nan],
        ['residuum', 'fuel_oil', np.nan],
        ], columns=['i', 'j', 'multiplier']).set_index(['i', 'j'])

    octane_data = pd.DataFrame([
        ['light_naphtha', 90],
        ['medium_naphtha', 80],
        ['heavy_naphtha', 70],
        ['reformed_gasoline', 115],
        ['cracked_gasoline', 105],
        ], columns=['i', 'octane']).set_index(['i'])

    petrol_data = pd.DataFrame([
        ['regular_petrol', 84],
        ['premium_petrol', 94],
        ], columns=['petrol', 'octane_lb']).set_index(['petrol'])

    vapour_pressure_data = pd.DataFrame([
        ['light_oil', 1.0],
        ['heavy_oil', 0.6],
        ['cracked_oil', 1.5],
        ['residuum', 0.05],
        ], columns=['oil', 'vapour_pressure']).set_index(['oil'])

    fuel_oil_ratio_data = pd.DataFrame([
        ['light_oil', 10],
        ['cracked_oil', 4],
        ['heavy_oil', 3],
        ['residuum', 1],
        ], columns=['oil', 'coefficient']).set_index(['oil'])

    final_product_data = pd.DataFrame([
        ['premium_petrol', 700],
        ['regular_petrol', 600],
        ['jet_fuel', 400],
        ['fuel_oil', 350],
        ['lube_oil', 150],
        ], columns=['product', 'profit']).set_index(['product'])

    vapour_pressure_ub = 1
    crude_total_ub = 45000
    naphtha_ub = 10000
    cracked_oil_ub = 8000
    lube_oil_lb = 500
    lube_oil_ub = 1000
    premium_ratio = 0.40

    ARCS = arc_data.index.tolist()
    arc_mult = arc_data['multiplier'].fillna(1)

    FINAL_PRODUCTS = final_product_data.index.tolist()
    final_product_data['profit'] = final_product_data['profit'] / 100
    profit = final_product_data['profit']

    ARCS = ARCS + [(i, 'sink') for i in FINAL_PRODUCTS]
    flow = m.add_variables(ARCS, name='flow', lb=0)
    NODES = np.unique([i for j in ARCS for i in j])

    m.set_objective(so.expr_sum(profit[i] * flow[i, 'sink']
                                 for i in FINAL_PRODUCTS
                                 if (i, 'sink') in ARCS),
                    name='totalProfit', sense=so.MAX)

    m.add_constraints((so.expr_sum(flow[a] for a in ARCS if a[0] == n) ==
                      so.expr_sum(arc_mult[a] * flow[a]
                                   for a in ARCS if a[1] == n)
                      for n in NODES if n not in ['source', 'sink']),
                      name='flow_balance')

    CRUDES = crude_data.index.tolist()
    crudeDistilled = m.add_variables(CRUDES, name='crudesDistilled', lb=0)
    crudeDistilled.set_bounds(ub=crude_data['crude_ub'])
    m.add_constraints((flow[i, j] == crudeDistilled[i]
                      for (i, j) in ARCS if i in CRUDES), name='distillation')

    OILS = ['light_oil', 'heavy_oil']
    CRACKED_OILS = [i+'_cracked' for i in OILS]
    oilCracked = m.add_variables(CRACKED_OILS, name='oilCracked', lb=0)
    m.add_constraints((flow[i, j] == oilCracked[i] for (i, j) in ARCS
                      if i in CRACKED_OILS), name='cracking')

    octane = octane_data['octane']
    PETROLS = petrol_data.index.tolist()
    octane_lb = petrol_data['octane_lb']
    vapour_pressure = vapour_pressure_data['vapour_pressure']

    m.add_constraints((so.expr_sum(octane[a[0]] * arc_mult[a] * flow[a]
                                    for a in ARCS if a[1] == p)
                       >= octane_lb[p] *
                      so.expr_sum(arc_mult[a] * flow[a]
                                   for a in ARCS if a[1] == p)
                      for p in PETROLS), name='blending_petrol')

    m.add_constraint(so.expr_sum(vapour_pressure[a[0]] * arc_mult[a] * flow[a]
                                  for a in ARCS if a[1] == 'jet_fuel') <=
                     vapour_pressure_ub *
                     so.expr_sum(arc_mult[a] * flow[a]
                                  for a in ARCS if a[1] == 'jet_fuel'),
                     name='blending_jet_fuel')

    fuel_oil_coefficient = fuel_oil_ratio_data['coefficient']
    sum_fuel_oil_coefficient = sum(fuel_oil_coefficient)
    m.add_constraints((sum_fuel_oil_coefficient * flow[a] ==
                      fuel_oil_coefficient[a[0]] * flow.sum('*', ['fuel_oil'])
                      for a in ARCS if a[1] == 'fuel_oil'),
                      name='blending_fuel_oil')

    m.add_constraint(crudeDistilled.sum('*') <= crude_total_ub,
                     name='crude_total_ub')

    m.add_constraint(so.expr_sum(flow[a] for a in ARCS
                                  if a[0].find('naphtha') > -1 and
                                  a[1] == 'reformed_gasoline')
                     <= naphtha_ub, name='naphtba_ub')

    m.add_constraint(so.expr_sum(flow[a] for a in ARCS if a[1] ==
                                  'cracked_oil') <=
                     cracked_oil_ub, name='cracked_oil_ub')

    m.add_constraint(flow['lube_oil', 'sink'] == [lube_oil_lb, lube_oil_ub],
                     name='lube_oil_range')

    m.add_constraint(flow.sum('premium_petrol', '*') >= premium_ratio *
                     flow.sum('regular_petrol', '*'), name='premium_ratio')

    res = m.solve(**kwargs)
    if res is not None:
        print(so.get_solution_table(crudeDistilled))
        print(so.get_solution_table(oilCracked))
        print(so.get_solution_table(flow))

        octane_sol = []
        for p in PETROLS:
            octane_sol.append(so.expr_sum(octane[a[0]] * arc_mult[a] *
                                           flow[a].get_value() for a in ARCS
                                           if a[1] == p) /
                              sum(arc_mult[a] * flow[a].get_value()
                                  for a in ARCS if a[1] == p))
        octane_sol = pd.Series(octane_sol, name='octane_sol', index=PETROLS)
        print(so.get_solution_table(octane_sol, octane_lb))
        print(so.get_solution_table(vapour_pressure))
        vapour_pressure_sol = sum(vapour_pressure[a[0]] *
                                  arc_mult[a] *
                                  flow[a].get_value() for a in ARCS
                                  if a[1] == 'jet_fuel') /\
            sum(arc_mult[a] * flow[a].get_value() for a in ARCS
                if a[1] == 'jet_fuel')
        print('Vapour_pressure_sol: {:.4f}'.format(vapour_pressure_sol))

        num_fuel_oil_ratio_sol = [arc_mult[a] * flow[a].get_value() /
                                  sum(arc_mult[b] *
                                      flow[b].get_value()
                                      for b in ARCS if b[1] == 'fuel_oil')
                                  for a in ARCS if a[1] == 'fuel_oil']
        num_fuel_oil_ratio_sol = pd.Series(num_fuel_oil_ratio_sol,
                                           name='num_fuel_oil_ratio_sol',
                                           index=[a[0] for a in ARCS
                                                  if a[1] == 'fuel_oil'])
        print(so.get_solution_table(fuel_oil_coefficient,
                                    num_fuel_oil_ratio_sol))

    return m.get_objective_value()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for COFOR loops.
"""

import os
import sys
import unittest
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../..')))
from util import assert_equal_wo_temps

from sasoptpy.actions import for_loop, cofor_loop, fix, solve, put_item


class TestCoforLoop(unittest.TestCase):
    """
    Unit tests for concurrent for (COFOR) statements
    """

    def setUp(self):
        so.reset()

    def test_simple_example(self):
        with so.Workspace('w') as w:
            x = so.VariableGroup(6, name='x', lb=0)
            so.Objective(
                so.expr_sum(x[i] for i in range(6)), name='z', sense=so.MIN)
            a1 = so.Constraint(x[1] + x[2] + x[3] <= 4, name='a1')

            for i in cofor_loop(so.exp_range(3, 6)):
                fix(x[1], i)
                solve()
                put_item(i, x[1], so.Symbol('_solution_status_'), names=True)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x {{0,1,2,3,4,5}} >= 0;
               min z = x[0] + x[1] + x[2] + x[3] + x[4] + x[5];
               con a1 : x[1] + x[2] + x[3] <= 4;
               cofor {TEMP1 in 3..5} do;
                  fix x[1]=TEMP1;
                  solve;
                  put TEMP1= x[1]= _solution_status_=;
               end;
            quit;'''))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for fix/unfix statements.
"""

import os
import sys
import unittest
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../..')))
from util import assert_equal_wo_temps

from sasoptpy.actions import fix, unfix

class TestFix(unittest.TestCase):
    """
    Unit tests for FIX and UNFIX statements
    """

    def setUp(self):
        so.reset()

    def test_regular_fix(self):

        from sasoptpy.actions import solve

        with so.Workspace('w') as w:
            x = so.Variable(name='x')
            fix(x, 1)
            solve()
            unfix(x)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x;
               fix x=1;
               solve;
               unfix x;
            quit;'''))

    def test_with_multiple_ops(self):

        from sasoptpy.actions import solve, cofor_loop

        with so.Workspace('w') as w:
            x = so.VariableGroup(4, name='x')
            for i in cofor_loop(range(4)):
                fix((x[0], i), (x[1], 1))
                solve()
                unfix(x[0], (x[1], 2))

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x {{0,1,2,3}};
               cofor {TEMP1 in 0..3} do;
                  fix x[0]=TEMP1 x[1]=1;
                  solve;
                  unfix x[0] x[1]=2;
               end;
            quit;'''))




import re
import sasoptpy

def parse_optmodel_table(table):
    parsed_df = table[['Label1', 'cValue1']]
    parsed_df.columns = ['Label', 'Value']
    parsed_df = parsed_df.set_index(['Label'])
    return parsed_df

def wrap_long_lines(code, max_length=30000):
    long_line_regex = r".{" + str(max_length) + r",}\n?"
    partition_regex = r"(?=.{" + str(max_length) + r",}\n?)(.{" + \
                      str(round(max_length / 3)) + r",}?)([\,\ ]+)(.+)"
    subst = "\\1\\2\\n\\3"

    hits = re.findall(long_line_regex, code)
    line_lengths = [len(i) for i in hits]
    while len(hits) > 0:
        code = re.sub(partition_regex, subst, code)
        hits = re.findall(long_line_regex, code)
        new_line_lengths = [len(i) for i in hits]
        if line_lengths == new_line_lengths:
            break
        else:
            line_lengths = new_line_lengths
    return code

def replace_long_names(code):
    conversion = dict()
    matches = re.findall(r'[a-zA-Z\_\d]{32,}', code)
    if len(matches) > 0:
        print('NOTE: Some object names are longer than 32 characters, '
              'they will be replaced when submitting')
        unique_matches = list(set(matches))
        for i in unique_matches:
            new_name = sasoptpy.util.get_next_name()
            conversion[new_name] = i
            code = re.sub(
                r'\b' + i + r'\b', new_name, code)

    return code, conversion




from math import inf

import numpy as np
import pandas as pd

import sasoptpy

def to_optmodel(caller, **kwargs):
    if sasoptpy.util.is_model(caller):
        return to_optmodel_for_solve(caller, **kwargs)
    elif sasoptpy.util.is_workspace(caller):
        return to_optmodel_for_session(caller, **kwargs)


def to_optmodel_for_solve(model, **kwargs):
    solve_option_keys = ('with', 'obj', 'objective', 'noobj', 'noobjective', 'relaxint', 'primalin', 'linearize')

    header = kwargs.get('header', True)
    ods = kwargs.get('ods', False)
    solve = kwargs.get('solve', True)
    options = kwargs.get('options', dict())
    primalin = kwargs.get('primalin', False)
    parse_results = kwargs.get('parse', False)
    multi_obj = False

    # Based on creation order
    s = ''
    if header:
        s = 'proc optmodel;\n'
    body = ''

    all_components_dict = {
        **model._setDict,
        **model._parameterDict,
        **model._statementDict,
        **model._variableDict,
        **model._impvarDict,
        **model._constraintDict,
        **model._objectiveDict,
    }

    all_components = list(all_components_dict.values()) + [model._objective]

    sorted_comp = sorted(all_components, key=lambda i: i._objorder)
    for cm in sorted_comp:
        if (sasoptpy.core.util.is_regular_component(cm)):
            body += cm._defn() + '\n'
            if hasattr(cm, '_member_defn'):
                mdefn = cm._member_defn()
                if mdefn != '':
                    body += mdefn + '\n'

    dropped_comps = ' '.join(list(model._get_dropped_vars().keys()) +
                             list(model._get_dropped_cons().keys()))
    if dropped_comps != '':
        body += mdefn + 'drop ' + dropped_comps + ';\n'

    # Solve block
    if solve:
        body += 'solve'
        pre_opts = []
        pos_opts = []

        if primalin:
            options['primalin'] = True

        if options:
            for key, value in options.items():
                if key in solve_option_keys:
                    if key == 'with':
                        pre_opts.append('with ' + options['with'])
                    elif key == 'relaxint' and options[key] is True:
                        pre_opts.append('relaxint')
                    elif key == 'obj' or key == 'objectives':
                        pre_opts.append('obj ({})'.format(' '.join(i.get_name() for i in options[key])))
                        multi_obj = True
                    elif key == 'primalin' and options[key] is True:
                        pos_opts.append('primalin')
                        primalin_set = True
                    elif key == 'linearize' and options[key] is True:
                        pre_opts.append('linearize')
                else:
                    if type(value) is dict:
                        pos_opts.append('{}=('.format(key) + ','.join(
                            '{}={}'.format(i, j)
                            for i, j in value.items()) + ')')
                    else:
                        pos_opts.append('{}={}'.format(key, value))

            if pre_opts != '':
                body += ' ' + ' '.join(pre_opts)
            if pos_opts != '':
                body += ' / ' + ' '.join(pos_opts)
        body += ';\n'

    # Output ODS tables
    if ods:
        body += 'ods output PrintTable=primal_out;\n'

    if parse_results:
        body += 'create data solution from [i]= {1.._NVAR_} var=_VAR_.name value=_VAR_ lb=_VAR_.lb ub=_VAR_.ub rc=_VAR_.rc;\n'

    if ods:
        body += 'ods output PrintTable=dual_out;\n'

    if parse_results:
        body += 'create data dual from [j] = {1.._NCON_} con=_CON_.name value=_CON_.body dual=_CON_.dual;\n'

    if multi_obj:
        body += 'create data allsols from [s]=(1.._NVAR_) name=_VAR_[s].name {j in 1.._NSOL_} <col(\'sol_\'||j)=_VAR_[s].sol[j]>;\n'

    # After-solve statements
    for i in model._postSolveDict.values():
        body += i._defn() + '\n'

    s += sasoptpy.util.addSpaces(body, 3)

    if header:
        s += 'quit;'
    return(s)


def to_optmodel_for_session(workspace, **kwargs):

    header = kwargs.get('header', True)
    parse = kwargs.get('parse', False)
    ods = kwargs.get('ods', False)


    s = ''

    if header:
        s += 'proc optmodel;\n'

    allcomp = workspace.get_elements()

    memberdefs = []
    for cm in allcomp:
        if (sasoptpy.core.util.is_regular_component(cm)):
            component_defn = cm._defn()
            if hasattr(cm, '_member_defn'):
                mdefn = cm._member_defn()
                if mdefn != '':
                    component_defn += '\n' + mdefn
            memberdefs.append(component_defn)

    memberdefs = '\n'.join(memberdefs)
    s += sasoptpy.util.addSpaces(memberdefs, 3)

    if parse:
        parse_str = '\ncreate data solution from [i]= {1.._NVAR_} var=_VAR_.name value=_VAR_ lb=_VAR_.lb ub=_VAR_.ub rc=_VAR_.rc;\n'
        parse_str += 'create data dual from [j] = {1.._NCON_} con=_CON_.name value=_CON_.body dual=_CON_.dual;'
        s += sasoptpy.util.addSpaces(parse_str, 3)

    if header:
        s += '\nquit;'
    return s



import sasoptpy as so
import pandas as pd
from sasoptpy.util import iterate, concat
from sasoptpy.actions import (
    read_data, create_data, cofor_loop, for_loop, solve, if_condition, diff,
    print_item, inline_condition)


def test(cas_conn, get_tables=False):

    input_list = pd.DataFrame(
        ['staff', 'showroom', 'pop1', 'pop2', 'alpha_enq', 'beta_enq'],
        columns=['input'])
    input_data = cas_conn.upload_frame(
        data=input_list, casout={'name': 'input_data', 'replace': True})

    output_list = pd.DataFrame(
        ['alpha_sales', 'beta_sales', 'profit'], columns=['output'])
    output_data = cas_conn.upload_frame(
        data=output_list, casout={'name': 'output_data', 'replace': True})

    problem_data = pd.DataFrame([
        ['Winchester', 7, 8, 10, 12, 8.5, 4, 2, 0.6, 1.5],
        ['Andover', 6, 6, 20, 30, 9, 4.5, 2.3, 0.7, 1.6],
        ['Basingstoke', 2, 3, 40, 40, 2, 1.5, 0.8, 0.25, 0.5],
        ['Poole', 14, 9, 20, 25, 10, 6, 2.6, 0.86, 1.9],
        ['Woking', 10, 9, 10, 10, 11, 5, 2.4, 1, 2],
        ['Newbury', 24, 15, 15, 13, 25, 19, 8, 2.6, 4.5],
        ['Portsmouth', 6, 7, 50, 40, 8.5, 3, 2.5, 0.9, 1.6],
        ['Alresford', 8, 7.5, 5, 8, 9, 4, 2.1, 0.85, 2],
        ['Salisbury', 5, 5, 10, 10, 5, 2.5, 2, 0.65, 0.9],
        ['Guildford', 8, 10, 30, 35, 9.5, 4.5, 2.05, 0.75, 1.7],
        ['Alton', 7, 8, 7, 8, 3, 2, 1.9, 0.7, 0.5],
        ['Weybridge', 5, 6.5, 9, 12, 8, 4.5, 1.8, 0.63, 1.4],
        ['Dorchester', 6, 7.5, 10, 10, 7.5, 4, 1.5, 0.45, 1.45],
        ['Bridport', 11, 8, 8, 10, 10, 6, 2.2, 0.65, 2.2],
        ['Weymouth', 4, 5, 10, 10, 7.5, 3.5, 1.8, 0.62, 1.6],
        ['Portland', 3, 3.5, 3, 2, 2, 1.5, 0.9, 0.35, 0.5],
        ['Chichester', 5, 5.5, 8, 10, 7, 3.5, 1.2, 0.45, 1.3],
        ['Petersfield', 21, 12, 6, 8, 15, 8, 6, 0.25, 2.9],
        ['Petworth', 6, 5.5, 2, 2, 8, 5, 1.5, 0.55, 1.55],
        ['Midhurst', 3, 3.6, 3, 3, 2.5, 1.5, 0.8, 0.2, 0.45],
        ['Reading', 30, 29, 120, 80, 35, 20, 7, 2.5, 8],
        ['Southampton', 25, 16, 110, 80, 27, 12, 6.5, 3.5, 5.4],
        ['Bournemouth', 19, 10, 90, 12, 25, 13, 5.5, 3.1, 4.5],
        ['Henley', 7, 6, 5, 7, 8.5, 4.5, 1.2, 0.48, 2],
        ['Maidenhead', 12, 8, 7, 10, 12, 7, 4.5, 2, 2.3],
        ['Fareham', 4, 6, 1, 1, 7.5, 3.5, 1.1, 0.48, 1.7],
        ['Romsey', 2, 2.5, 1, 1, 2.5, 1, 0.4, 0.1, 0.55],
        ['Ringwood', 2, 3.5, 2, 2, 1.9, 1.2, 0.3, 0.09, 0.4],
    ], columns=['garage_name', 'staff', 'showroom', 'pop1', 'pop2', 'alpha_enq',
                'beta_enq', 'alpha_sales', 'beta_sales', 'profit'])
    garage_data = cas_conn.upload_frame(
        data=problem_data, casout={'name': 'garage_data', 'replace': True})

    with so.Workspace(name='efficiency_analysis', session=cas_conn) as w:
        inputs = so.Set(name='INPUTS', settype=so.string)
        read_data(table=input_data, index={'target': inputs, 'key': 'input'})

        outputs = so.Set(name='OUTPUTS', settype=so.string)
        read_data(table=output_data, index={'target': outputs, 'key': 'output'})

        garages = so.Set(name='GARAGES', settype=so.number)
        garage_name = so.ParameterGroup(garages, name='garage_name', ptype=so.string)
        input = so.ParameterGroup(inputs, garages, name='input')
        output = so.ParameterGroup(outputs, garages, name='output')
        r = read_data(table=garage_data, index={'target': garages, 'key': so.N},
                      columns=[garage_name])
        with iterate(inputs, 'i') as i:
            r.append({'index': i, 'target': input[i, so.N], 'column': i})
        with iterate(outputs, 'i') as i:
            r.append({'index': i, 'target': output[i, so.N], 'column': i})

        k = so.Parameter(name='k', ptype=so.number)
        efficiency_number = so.ParameterGroup(garages, name='efficiency_number')
        weight_sol = so.ParameterGroup(garages, garages, name='weight_sol')

        weight = so.VariableGroup(garages, name='Weight', lb=0)
        inefficiency = so.Variable(name='Inefficiency', lb=0)

        obj = so.Objective(inefficiency, name='Objective', sense=so.maximize)

        input_con = so.ConstraintGroup(
            (so.expr_sum(input[i, j] * weight[j] for j in garages) <= input[i, k]
             for i in inputs), name='input_con')
        output_con = so.ConstraintGroup(
            (so.expr_sum(output[i, j] * weight[j] for j in garages) >= output[i, k] * inefficiency
             for i in outputs), name='output_con')

        for kk in cofor_loop(garages):
            k.set_value(kk)
            solve()
            efficiency_number[k] = 1 / inefficiency.sol
            for j in for_loop(garages):
                def if_block():
                    weight_sol[k, j] = weight[j].sol
                def else_block():
                    weight_sol[k, j] = None
                if_condition(weight[j].sol > 1e-6, if_block, else_block)

        efficient_garages = so.Set(
            name='EFFICIENT_GARAGES',
            value=[j.sym for j in garages if j.sym.under_condition(efficiency_number[j] >= 1)])
        inefficient_garages = so.Set(value=diff(garages, efficient_garages), name='INEFFICIENT_GARAGES')

        p1 = print_item(garage_name, efficiency_number)
        ed = create_data(table='efficiency_data', index={'key': ['garage']}, columns=[
            garage_name, efficiency_number
        ])
        with iterate(inefficient_garages, 'inefficient_garage') as i:
            wd = create_data(table='weight_data_dense',
                             index={'key': [i], 'set': [i.get_set()]},
                             columns=[garage_name, efficiency_number])
            with iterate(efficient_garages, 'efficient_garage') as j:
                wd.append({
                    'name': concat('w', j),
                    'expression': weight_sol[i, j],
                    'index': j
                })

        filtered_set = so.InlineSet(
            lambda: ((g1, g2)
                     for g1 in inefficient_garages
                     for g2 in efficient_garages
                     if inline_condition(weight_sol[g1, g2] != None)))
        wds = create_data(table='weight_data_sparse',
                          index={'key': ['i', 'j'], 'set': [filtered_set]},
                          columns=[weight_sol])

    print(w.to_optmodel())
    w.submit()

    print('Print Table:')
    print(p1.get_response())

    print('Efficiency Data:')
    print(ed.get_response())

    print('Weight Data (Dense):')
    print(wd.get_response())

    print('Weight Data (Sparse):')
    print(wds.get_response())

    if get_tables:
        return obj.get_value(), ed.get_response()
    else:
        return obj.get_value()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from math import inf

import sasoptpy


class Config:
    """
    Option manager for sasoptpy
    """

    def __init__(self, *args):
        self.defaults = _get_default_config()
        self.overridden = dict()

        if len(args) > 1 and len(args) % 2 != 1:
            keys = args[::2]
            values = args[1::2]
            self.overridden = dict(zip(keys, values))

    def __getitem__(self, key):
        if key in self.overridden:
            return self.overridden[key]
        return self.defaults.get(key, None)

    def __setitem__(self, key, value):
        self.overridden[key] = value

    def __delitem__(self, key):
        if key in self.overridden:
            del self.overridden[key]
        else:
            raise KeyError('ERROR: Invalid config key: {}'.format(key))

    @property
    def keys(self):
        dicts = [self.defaults, self.overridden]
        merged_keys = set().union(*dicts)
        return list(merged_keys)

    def __iter__(self):
        for key in self.keys:
            yield key

    def reset(self):
        self.overridden = dict()


def _get_default_config():
    config = dict()
    config['verbosity'] = 3
    config['max_digits'] = 12
    config['print_digits'] = 6
    config['valid_outcomes'] = ['OPTIMAL', 'ABSFCONV', 'BEST_FEASIBLE']
    config['default_sense'] = sasoptpy.MIN
    config['submit_realtime'] = True
    config['default_bounds'] = {
        sasoptpy.CONT: {'lb': -inf, 'ub': inf},
        sasoptpy.INT: {'lb': -inf, 'ub': inf},
        sasoptpy.BIN: {'lb': 0, 'ub': 1}
    }
    config['generic_naming'] = False

    return config


def _load_default_config():
    sasoptpy.config = Config()
    sasoptpy.default_config_keys = sasoptpy.config.keys




from contextlib import contextmanager

import sasoptpy
from sasoptpy.util import register_to_function_container


def register_actions():
    register_to_function_container(
        read_data, sasoptpy.abstract.statement.ReadDataStatement.read_data)
    register_to_function_container(
        create_data, sasoptpy.abstract.statement.CreateDataStatement.create_data)
    register_to_function_container(
        solve, sasoptpy.abstract.statement.SolveStatement.solve)
    register_to_function_container(
        for_loop, sasoptpy.abstract.statement.ForLoopStatement.for_loop)
    register_to_function_container(
        cofor_loop, sasoptpy.abstract.statement.CoForLoopStatement.cofor_loop)
    register_to_function_container(
        if_condition, sasoptpy.abstract.statement.IfElseStatement.if_condition)
    register_to_function_container(
        switch_conditions, sasoptpy.abstract.statement.if_else.SwitchStatement.switch_condition)
    register_to_function_container(
        set_value, sasoptpy.abstract.statement.Assignment.set_value)
    register_to_function_container(
        fix, sasoptpy.abstract.statement.FixStatement.fix)
    register_to_function_container(
        unfix, sasoptpy.abstract.statement.UnfixStatement.unfix)
    register_to_function_container(
        set_objective, sasoptpy.abstract.statement.ObjectiveStatement.set_objective)
    register_to_function_container(
        print_item, sasoptpy.abstract.statement.PrintStatement.print_item)
    register_to_function_container(
        expand, sasoptpy.abstract.statement.LiteralStatement.expand)
    register_to_function_container(
        drop, sasoptpy.abstract.statement.DropStatement.drop_constraint)
    register_to_function_container(
        restore, sasoptpy.abstract.statement.RestoreStatement.restore_constraint)
    register_to_function_container(
        put_item, sasoptpy.abstract.statement.PrintStatement.put_item)
    register_to_function_container(
        union, sasoptpy.abstract.statement.LiteralStatement.union)
    register_to_function_container(
        diff, sasoptpy.abstract.statement.LiteralStatement.diff)
    register_to_function_container(
        substring, sasoptpy.abstract.statement.LiteralStatement.substring)
    register_to_function_container(
        use_problem, sasoptpy.abstract.statement.LiteralStatement.use_problem)


@sasoptpy.containable
def read_data(table, index, columns):
    """
    Reads data tables inside Set and Parameter objects

    Parameters
    ----------
    table : string or :class:`swat.cas.table.CASTable`
        Table object or name to be read, case-insensitive
    index : dict
        Index properties of the table

        Has two main members:

        - target : :class:`sasoptpy.abstract.Set`
            Target Set object to be read into
        - key : string, list or None
            Column name to be be read from.

            For multiple indices, key should be a list of string or
            :class:`sasoptpy.abstract.SetIterator` objects

        For a given set `YEAR` and column name `year_no`, the index dictionary
        should be written as:

        >>> {'target': YEAR, 'key': 'year_no'}

        If index is simply the row number, use `'key': so.N` which is equivalent
        to the special `_N_` character in the SAS language.
    columns : list
        A list of dictionaries, each holding column properties.

        Columns are printed in the specified order. Each column should be
        represented as a dictionary with following fields:

        - target : :class:`sasoptpy.abstract.ParameterGroup`
            Target parameter object to be read into
        - column : string
            Column name to be read from
        - index : :class:`sasoptpy.SetIterator`, optional
            Subindex for specific column, needed for complex operations

        If the name of the :class:`sasoptpy.abstract.Parameter` object is same
        as the column name, the following call is enough:

        >>> p = so.Parameter(name='price')
        >>> read_data(..., columns=[{'target': p}])

        For reading a different column name, `column` field should be
        specified:

        >>> {'target': p, 'column': 'price_usd'}

        When working with :class:`ParameterGroup` objects, sometimes a secondary
        loop is needed. This is achieved by using the `index` field, along with
        the :meth:`sasoptpy.abstract.statement.ReadDataStatement.append` method.

    Returns
    -------
    r : :class:`sasoptpy.abstract.statement.ReadDataStatement`
        Read data statement object, which includes all properties

        Additional columns can be added using the
        :meth:`sasoptpy.abstract.statement.ReadDataStatement.append`
        function.

    Examples
    --------

    Reading a regular set:

    >>> with Workspace('test_workspace') as ws:
    >>>     ITEMS = Set(name='ITEMS')
    >>>     value = ParameterGroup(ITEMS, name='value', init=0)
    >>>     get = VariableGroup(ITEMS, name='get', vartype=so.INT, lb=0)
    >>>     read_data(
    ...         table="values",
    ...         index={'target': ITEMS, 'key': None},
    ...         columns=[{'target': value}])
    >>> print(so.to_optmodel(w))
    proc optmodel;
        set ITEMS;
        num value {ITEMS} init 0;
        var get {{ITEMS}} integer >= 0;
        read data values into ITEMS value;
    quit;

    Reading with row index:

    >>> with so.Workspace('test_read_data_n') as ws:
    >>>     ASSETS = so.Set(name='ASSETS')
    >>>     ret = so.ParameterGroup(ASSETS, name='return', ptype=so.NUM)
    >>>     read_data(
    ...         table='means',
    ...         index={'target': ASSETS, 'key': so.N},
    ...         columns=[{'target': ret}]
    ...     )
    >>> print(so.to_optmodel(w))
    proc optmodel;
        set ASSETS;
        num return {ASSETS};
        read data means into ASSETS=[_N_] return;
    quit;

    Reading with no index set and subindex:

    >>> with so.Workspace('test_read_data_no_index_expression') as ws:
    >>>     ASSETS = so.Set(name='ASSETS')
    >>>     cov = so.ParameterGroup(ASSETS, ASSETS, name='cov', init=0)
    >>>     with iterate(ASSETS, 'asset1') as asset1, iterate(ASSETS, 'asset2') as asset2:
    >>>         read_data(
    ...             table='covdata',
    ...             index={'key': [asset1, asset2]},
    ...             columns=[
    ...                 {'target': cov},
    ...                 {'target': cov[asset2, asset1],
    ...                 'column': 'cov'}])
    >>> print(so.to_optmodel(w))
    proc optmodel;
        set ASSETS;
        num cov {ASSETS, ASSETS} init 0;
        read data covdata into [asset1 asset2] cov cov[asset2, asset1]=cov;
    quit;

    Reading a column with multiple indices:

    >>> with so.Workspace(name='test_read_data_idx_col') as ws:
    >>>     dow = so.Set(name='DOW', value=so.exp_range(1, 6))
    >>>     locs = so.Set(name='LOCS', settype=so.STR)
    >>>     demand = so.ParameterGroup(locs, dow, name='demand')
    >>>     with iterate(locs, name='loc') as loc:
    >>>         r = read_data(
    ...             table='dmnd',
    ...             index={'target': locs, 'key': loc}
    ...         )
    >>>         with iterate(dow, name='d') as d:
    >>>             r.append({
    ...                 'index': d,
    ...                 'target': demand[loc, d],
    ...                 'column': concat('day', d)
    ...             })
    >>> optmodel_code = so.to_optmodel(ws)
    proc optmodel;
        set DOW = 1..5;
        set <str> LOCS;
        num demand {LOCS, DOW};
        read data dmnd into LOCS=[loc] {d in DOW} < demand[loc, d]=col('day' || d) >;
    quit;

    See also
    --------
    :class:`tests.abstract.statement.test_read_data.TestReadData`

    """
    return sasoptpy.abstract.ReadDataStatement(table, index, columns)


@sasoptpy.containable(standalone=False)
def create_data(table, index, columns):
    """
    Creates data tables from variables, parameters, and expressions

    Parameters
    ----------
    table : string
        Name of the table to be created
    index : dict
        Table index properties

        This dictionary can be empty if no index is needed. It can have
        following fields:

        - key : list
            List of index keys. Keys can be string or
            :class:`sasoptpy.abstract.SetIterator` objects
        - set : list
            List of sets that is being assigned to keys
    columns : list
        List of columns. Columns can be :class:`sasoptpy.abstract.Parameter`,
        :class:`sasoptpy.abstract.ParameterGroup` objects or dictionaries. If
        specified as a dictionary, each can have following keys:

        - name : string
            Name of the column in output table
        - expression : :class:`sasoptpy.core.Expression`
            Any expression
        - index : list or :class:`sasoptpy.abstract.SetIterator`
            Index for internal loops

        The `index` field can be used when a subindex is needed. When
        specified as a list, members should be
        :class:`sasoptpy.abstract.SetIterator` objects.
        See examples for more details.

    Examples
    --------

    Regular column:

    >>> with so.Workspace('w') as w:
    >>>     m = so.Parameter(name='m', value=7)
    >>>     n = so.Parameter(name='n', value=5)
    >>>     create_data(table='example', index={}, columns=[m, n])
    >>> print(so.to_optmodel(w))
    proc optmodel;
        num m = 7;
        num n = 5;
        create data example from m n;
    quit;

    Column with name:

    >>> with so.Workspace('w') as w:
    >>>     m = so.Parameter(name='m', value=7)
    >>>     n = so.Parameter(name='n', value=5)
    >>>     create_data(table='example', index={}, columns=[
    ...         {'name': 'ratio', 'expression': m/n}
    ...     ])
    >>> print(so.to_optmodel(w))
    proc optmodel;
        num m = 7;
        num n = 5;
        create data example from ratio=((m) / (n));
    quit;

    Column name with concat:

    >>> from sasoptpy.util import concat
    >>> with so.Workspace('w') as w:
    >>>     m = so.Parameter(name='m', value=7)
    >>>     n = so.Parameter(name='n', value=5)
    >>>     create_data(table='example', index={}, columns=[
    ...         {'name': concat('s', n), 'expression': m+n}
    ...     ])
    >>> print(so.to_optmodel(w))
    proc optmodel;
        num m = 7;
        num n = 5;
        create data example from col('s' || n)=(m + n);
    quit;

    Table with index:

    >>> with so.Workspace('w') as w:
    >>>     m = so.ParameterGroup(
    >>>         so.exp_range(1, 6), so.exp_range(1, 4), name='m', init=0)
    >>>     m[1, 1] = 1
    >>>     m[4, 1] = 1
    >>>     S = so.Set(name='ISET', value=[i**2 for i in range(1, 3)])
    >>>     create_data(
    ...         table='example',
    ...         index={'key': ['i', 'j'], 'set': [S, [1, 2]]},
    ...         columns=[m]
    ...     )
    >>> print(so.to_optmodel(w))
    proc optmodel;
        num m {1..5, 1..3} init 0;
        m[1, 1] = 1;
        m[4, 1] = 1;
        set ISET = {1,4};
        create data example from [i j] = {{ISET,{1,2}}} m;
    quit;

    Index over Python range:

    >>> with so.Workspace('w') as w:
    >>>     s = so.Set(name='S', value=so.exp_range(1, 6))
    >>>     x = so.VariableGroup(s, name='x')
    >>>     x[1] = 1
    >>>     create_data(table='example',
    ...         index={'key': ['i'], 'set': so.exp_range(1, 4)}, columns=[x])
    >>> print(so.to_optmodel(w))
    proc optmodel;
        set S = 1..5;
        var x {{S}};
        x[1] = 1;
        create data example from [i] = {1..3} x;
    quit;

    Append column with index:

    >>> from sasoptpy.util import iterate, concat
    >>> with so.Workspace('w', session=session) as w:
    >>>     alph = so.Set(name='alph', settype=so.string, value=['a', 'b', 'c'])
    >>>     x = so.VariableGroup([1, 2, 3], alph, name='x', init=2)
    >>>     with iterate(so.exp_range(1, 4), name='i') as i:
    >>>         c = create_data(
    ...             table='example',
    ...             index={'key': [i], 'set': [i.get_set()]})
    >>>         with iterate(alph, name='j') as j:
    >>>             c.append(
    ...                 {'name': concat('x', j),
    ...                  'expression': x[i, j],
    ...                  'index': j})
    >>> print(so.to_optmodel(w))
    proc optmodel;
        set <str> alph = {'a','b','c'};
        var x {{1,2,3}, {alph}} init 2;
        create data example from [i] = {{1..3}} {j in alph} < col('x' || j)=(x[i, j]) >;
    quit;

    Multiple column indices:

    >>> from sasoptpy.util import concat, iterate
    >>> with so.Workspace('w') as w:
    >>>     S = so.Set(name='S', value=[1, 2, 3])
    >>>     T = so.Set(name='T', value=[1, 3, 5])
    >>>     x = so.VariableGroup(S, T, name='x', init=1)
    >>>     with iterate(S, name='i') as i, iterate(T, name='j') as j:
    >>>         create_data(
    ...             table='out',
    ...             index={},
    ...             columns=[
    ...                 {'name': concat('x', concat(i, j)), 'expression': x[i, j],
    ...                  'index': [i, j]}])
    >>> print(so.to_optmodel(w))
    proc optmodel;
        set S = {1,2,3};
        set T = {1,3,5};
        var x {{S}, {T}} init 1;
        create data out from {i in S, j in T} < col('x' || i || j)=(x[i, j]) >;
    quit;

    See also
    --------
    :class:`tests.abstract.statement.test_create_data.TestCreateData`

    """
    pass


@sasoptpy.containable(standalone=False)
def solve(options=None, primalin=False):
    """
    Solves the active optimization problem and generates results

    Parameters
    ----------
    options : dict, optional
        Solver options

        This dictionary can have several fields.

        - with : string
            Name of the solver, see possible values under Notes.

        See :ref:`solver-options` for a list of solver options. All fields in
        options (except `with`) is passed directly to the solver.

    primalin : bool, optional
        When set to `True`, uses existing variable values as an initial point
        in MILP solver

    Notes
    -----

    Possible solver names for `with` parameter:

    * `lp` : Linear programming
    * `milp` : Mixed integer linear programming
    * `nlp` : General nonlinear programming
    * `qp` : Quadratic programming
    * `blackbox` : Black-box optimization

    SAS Optimization also has a constraint programming solver (clp), and network
    solver (network) but they are not currently supported by sasoptpy.

    Returns
    -------
    ss : :class:`sasoptpy.abstract.statement.SolveStatement`
        Solve statement object.

        Contents of the response can be retrieved using `get_response` function.

    Examples
    --------

    Regular solve:

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x', lb=1, ub=10)
    >>>     o = so.Objective(2*x, sense=so.maximize, name='obj')
    >>>     s = solve()
    >>>     p = print_item(x)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x >= 1 <= 10;
        max obj = 2 * x;
        solve;
        print x;
    quit;

    Option alternatives:

    >>> with so.Workspace('w') as w:
    >>>     # Problem declaration, etc..
    >>>     solve()
    >>>     solve(options={'with': 'milp'})
    >>>     solve(options={'with': 'milp'}, primalin=True)
    >>>     solve(options={'with': 'milp', 'presolver': None, 'feastol': 1e-6,
    >>>                    'logfreq': 2, 'maxsols': 3, 'scale': 'automatic',
    >>>                    'restarts': None, 'cutmir': 'aggressive'})
    >>> print(so.to_optmodel(w))
    proc optmodel;
        solve;
        solve with milp;
        solve with milp / primalin;
        solve with milp / presolver=None feastol=1e-06 logfreq=2 maxsols=3 scale=automatic restarts=None cutmir=aggressive;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def for_loop(*args):
    """
    Creates a for-loop container to be executed on the server

    Parameters
    ----------
    args : :class:`sasoptpy.abstract.Set` objects
        Any number of :class:`sasoptpy.abstract.Set` objects can be given

    Returns
    -------
    set_iterator : :class:`sasoptpy.abstract.SetIterator`, :class:`sasoptpy.abstract.SetIteratorGroup`
        Set iterators to be used inside for-loop

    Examples
    --------

    Regular for loop:

    >>> with so.Workspace('w') as w:
    >>>     r = so.exp_range(1, 11)
    >>>     x = so.VariableGroup(r, name='x')
    >>>     for i in for_loop(r):
    >>>         x[i] = 1
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x {{1,2,3,4,5,6,7,8,9,10}};
        for {o13 in 1..10} do;
            x[o13] = 1;
        end;
    quit;

    Nested for loops:

    >>> from sasoptpy.actions import put_item
    >>> with so.Workspace('w') as w:
    >>>     for i in for_loop(range(1, 3)):
    >>>         for j in for_loop(['a', 'b']):
    >>>             put_item(i, j)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        for {o2 in 1..2} do;
            for {o5 in {'a','b'}} do;
                put o2 o5;
            end;
        end;
    quit;

    Multiple set for-loops:

    >>> with so.Workspace('w') as w:
    >>>     r = so.Set(name='R', value=range(1, 11))
    >>>     c = so.Set(name='C', value=range(1, 6))
    >>>     a = so.ParameterGroup(r, c, name='A', ptype=so.number)
    >>>     for (i, j) in for_loop(r, c):
    >>>         a[i, j] = 1
    >>> print(so.to_optmodel(w))
    proc optmodel;
        set R = 1..10;
        set C = 1..5;
        num A {R, C};
        for {o5 in R, o7 in C} do;
            A[o5, o7] = 1;
        end;
    quit;

    See also
    --------
    :func:`sasoptpy.actions.cofor_loop`

    Notes
    -----

    For tasks that can be run concurrently, consider using
    :func:`sasoptpy.actions.cofor_loop`

    """
    pass


@sasoptpy.containable(standalone=False)
def cofor_loop(*args):
    """
    Creates a cofor-loop to be executed on the server concurrently

    Parameters
    ----------
    args : :class:`sasoptpy.abstract.Set` objects
        Any number of :class:`sasoptpy.abstract.Set` objects can be specified

    Returns
    -------
    set_iterator : :class:`sasoptpy.abstract.SetIterator`, :class:`sasoptpy.abstract.SetIteratorGroup`
        Set iterators to be used inside cofor-loop

    Examples
    --------

    >>> with so.Workspace('w') as w:
    >>>     x = so.VariableGroup(6, name='x', lb=0)
    >>>     so.Objective(
    >>>         so.expr_sum(x[i] for i in range(6)), name='z', sense=so.MIN)
    >>>     a1 = so.Constraint(x[1] + x[2] + x[3] <= 4, name='a1')
    >>>     for i in cofor_loop(so.exp_range(3, 6)):
    >>>         fix(x[1], i)
    >>>         solve()
    >>>         put_item(i, x[1], so.Symbol('_solution_status_'), names=True)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x {{0,1,2,3,4,5}} >= 0;
        min z = x[0] + x[1] + x[2] + x[3] + x[4] + x[5];
        con a1 : x[1] + x[2] + x[3] <= 4;
        cofor {o13 in 3..5} do;
            fix x[1]=o13;
            solve;
            put o13= x[1]= _solution_status_=;
        end;
    quit;

    See also
    --------
    :func:`sasoptpy.actions.for_loop`

    Notes
    -----

    A cofor-loop runs its content concurrently. For tasks that depend on each
    other, consider using :func:`sasoptpy.actions.for_loop`
    """
    pass


@sasoptpy.containable(standalone=False)
def if_condition(logic_expression, if_statement, else_statement=None):
    """
    Creates an if-else block

    Parameters
    ----------
    logic_expression : :class:`sasoptpy.Constraint` or :class:`sasoptpy.abstract.Condition`
        Logical condition for the True case

        For the condition, it is possible to combine constraints, such as

        >>> a = so.Parameter(value=5)
        >>> if_condition((a < 3) | (a > 6), func1, func2)

        Constraints should be combined using bitwise operators
        (& for `and`, | for `or`).
    if_statement : function or :class:`IfElseStatement`
        Python function or if-else statement to be called if the condition is True
    else_statement : function or :class:`IfElseStatement`, optional
        Python function or if-else statement to be called if the condition is False

    Examples
    --------

    Regular condition:

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x')
    >>>     x.set_value(0.5)
    >>>     def func1():
    >>>         x.set_value(1)
    >>>     def func2():
    >>>         x.set_value(0)
    >>>     if_condition(x > 1e-6, func1, func2)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x;
        x = 0.5;
        if x > 1e-06 then do;
            x = 1;
        end;
        else do;
            x = 0;
        end;
    quit;

    Combined conditions:

    >>> with so.Workspace('w') as w:
    >>>     p = so.Parameter(name='p')
    >>>     def case1():
    >>>         p.set_value(10)
    >>>     def case2():
    >>>         p.set_value(20)
    >>>     r = so.Parameter(name='r', value=10)
    >>>     if_condition((r < 5) | (r > 10), case1, case2)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        num p;
        num r = 10;
        if (r < 5) or (r > 10) then do;
            p = 10;
        end;
        else do;
            p = 20;
        end;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def switch_conditions(**args):
    """
    Creates several if-else blocks by using the specified arguments

    Parameters
    ----------
    args :
        Several arguments can be passed to the function

        Each case should follow a condition.
        You can use :class:`sasoptpy.Constraint` objects as conditions, and
        Python functions for the cases.

    Examples
    --------

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x')
    >>>     p = so.Parameter(name='p')
    >>>     x.set_value(2.5)
    >>>     def func1():
    >>>         p.set_value(1)
    >>>     def func2():
    >>>         p.set_value(2)
    >>>     def func3():
    >>>         p.set_value(3)
    >>>     def func4():
    >>>         p.set_value(0)
    >>>     switch_conditions(x < 1, func1, x < 2, func2, x < 3, func3, func4)
    >>> print(to.optmodel(w))
    proc optmodel;
        var x;
        num p;
        x = 2.5;
        if x < 1 then do;
            p = 1;
        end;
        else if x < 2 then do;
            p = 2;
        end;
        else if x < 3 then do;
            p = 3;
        end;
        else do;
            p = 0;
        end;
    quit;
    """
    pass


@sasoptpy.containable(standalone=False)
def set_value(left, right):
    """
    Creates an assignment statement

    Parameters
    ----------
    left : :class:`sasoptpy.Expression`
        Any expression (variable or parameter)
    right : :class:`sasoptpy.Expression` or float
        Right-hand-side expression

    Examples
    --------

    >>> with so.Workspace('ex_9_1_matirx_sqrt', session=None) as w:
    >>>     so.LiteralStatement('call streaminit(1);')
    >>>     n = so.Parameter(name='n', value=5)
    >>>     rn = so.Set(name='RN', value=so.exp_range(1, n))
    >>>     A = so.ParameterGroup(rn, rn, name='A', value="10-20*rand('UNIFORM')")
    >>>     P = so.ParameterGroup(rn, rn, name='P')
    >>>     for i in for_loop(rn):
    >>>         for j in for_loop(so.exp_range(i, n)):
    >>>             set_value(P[i, j], so.expr_sum(A[i, k] * A[j, k] for k in rn))
    >>> print(so.to_optmodel(w))
    proc optmodel;
        call streaminit(1);
        num n = 5;
        set RN = 1..n;
        num A {RN, RN} = 10-20*rand('UNIFORM');
        num P {RN, RN};
        for {o7 in RN} do;
            for {o10 in o7..n} do;
                P[o7, o10] = sum {k in RN} (A[o7, k] * A[o10, k]);
            end;
        end;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def fix(*args):
    """
    Fixes values of variables to the specified values

    Parameters
    ----------
    args : :class:`sasoptpy.Variable`, float, :class:`sasoptpy.Expression`, tuple
        Set of arguments to be fixed

        Arguments get paired (if not given in tuples) to allow several fix
        operations

    Examples
    --------

    Regular fix statement:

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x')
    >>>     fix(x, 1)
    >>>     solve()
    >>>     unfix(x)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x;
        fix x=1;
        solve;
        unfix x;
    quit;

    Multiple fix-unfix:

    >>> with so.Workspace('w') as w:
    >>>     x = so.VariableGroup(4, name='x')
    >>>     for i in cofor_loop(range(4)):
    >>>         fix((x[0], i), (x[1], 1))
    >>>         solve()
    >>>         unfix(x[0], (x[1], 2))
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x {{0,1,2,3}};
        cofor {o7 in 0..3} do;
            fix x[0]=o7 x[1]=1;
            solve;
            unfix x[0] x[1]=2;
        end;
    quit;

    See also
    --------
    :func:`sasoptpy.actions.unfix`,
    :class:`tests.abstract.statement.test_fix_unfix.TestFix`

    """
    pass


@sasoptpy.containable(standalone=False)
def unfix(*args):
    """
    Unfixes values of variables

    Parameters
    ----------
    args : :class:`sasoptpy.Variable` objects
        Set of arguments to be unfixed

    Examples
    --------

    Regular unfix statement:

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x')
    >>>     fix(x, 1)
    >>>     solve()
    >>>     unfix(x)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x;
        fix x=1;
        solve;
        unfix x;
    quit;

    Multiple fix-unfix:

    >>> with so.Workspace('w') as w:
    >>>     x = so.VariableGroup(4, name='x')
    >>>     for i in cofor_loop(range(4)):
    >>>         fix((x[0], i), (x[1], 1))
    >>>         solve()
    >>>         unfix(x[0], (x[1], 2))
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x {{0,1,2,3}};
        cofor {o7 in 0..3} do;
            fix x[0]=o7 x[1]=1;
            solve;
            unfix x[0] x[1]=2;
        end;
    quit;

    See also
    --------
    :func:`sasoptpy.actions.fix`,
    :class:`tests.abstract.statement.test_fix_unfix.TestFix`

    """
    pass


@sasoptpy.containable(standalone=False)
def set_objective(expression, name, sense):
    """
    Specifies the objective function

    Parameters
    ----------
    expression : :class:`sasoptpy.Expression`
        Objective function
    name : string
        Name of the objective function
    sense : string
        Direction of the objective function, `so.MAX` or `so.MIN`

    Examples
    --------

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x', lb=1)
    >>>     set_objective(x ** 3, name='xcube', sense=so.minimize)
    >>>     solve()
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x >= 1;
        MIN xcube = (x) ^ (3);
        solve;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def print_item(*args):
    """
    Prints the specified argument list on server

    Parameters
    ----------
    args : :class:`sasoptpy.Variable`, :class:`sasoptpy.Expression`
        Arbitrary number of arguments to be printed

        These values are printed on the server, but can be retrieved after
        execution

    Returns
    -------
    ps : :class:`sasoptpy.abstract.statement.PrintStatement`
        Print statement object.

        Contents of the response can be retrieved using `get_response` function.

    Examples
    --------

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x', lb=1, ub=10)
    >>>     o = so.Objective(2*x, sense=so.maximize, name='obj')
    >>>     s = solve()
    >>>     p = print_item(x)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x >= 1 <= 10;
        max obj = 2 * x;
        solve;
        print x;
    quit;
    >>> print(p.get_response())
          x
    0  10.0

    """
    pass


@sasoptpy.containable(standalone=False)
def put_item(*args, names=None):
    """
    Prints the specified item values to the output log

    Parameters
    ----------
    args : :class:`sasoptpy.Expression`, string
        Arbitrary elements to be put into log

        Variables, variable groups, and expressions can be printed to log
    names : bool, optional
        When set to `True`, prints the name of the arguments in the log

    Examples
    --------

    Regular operation:

    >>> with so.Workspace('w') as w:
    >>>     for i in for_loop(range(1, 3)):
    >>>         for j in for_loop(['a', 'b']):
    >>>             put_item(i, j)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        for {o2 in 1..2} do;
            for {o5 in {'a','b'}} do;
                put o2 o5;
            end;
        end;
    quit;

    Print with names:

    >>> with so.Workspace('w') as w:
    >>>     x = so.VariableGroup(6, name='x', lb=0)
    >>>     so.Objective(
    >>>         so.expr_sum(x[i] for i in range(6)), name='z', sense=so.MIN)
    >>>     a1 = so.Constraint(x[1] + x[2] + x[3] <= 4, name='a1')
    >>>     for i in cofor_loop(so.exp_range(3, 6)):
    >>>         fix(x[1], i)
    >>>         solve()
    >>>         put_item(i, x[1], so.Symbol('_solution_status_'), names=True)
    proc optmodel;
        var x {{0,1,2,3,4,5}} >= 0;
        min z = x[0] + x[1] + x[2] + x[3] + x[4] + x[5];
        con a1 : x[1] + x[2] + x[3] <= 4;
        cofor {o13 in 3..5} do;
            fix x[1]=o13;
            solve;
            put o13= x[1]= _solution_status_=;
        end;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def expand():
    """
    Prints expanded problem to output

    Examples
    --------
    >>> with so.Workspace(name='w') as w:
    >>>     x = so.VariableGroup(3, name='x')
    >>>     self.assertEqual(x[0].sym.get_conditions_str(), '')
    >>>     # solve
    >>>     x[0].set_value(1)
    >>>     x[1].set_value(5)
    >>>     x[2].set_value(0)
    >>>     c = so.ConstraintGroup(None, name='c')
    >>>     with iterate([0, 1, 2], 's') as i:
    >>>         with condition(x[i].sym > 0):
    >>>             c[i] = x[i] >= 1
    >>>     set_objective(x[0], name='obj', sense=so.MIN)
    >>>     expand()
    >>>     solve()
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x {{0,1,2}};
        x[0] = 1;
        x[1] = 5;
        x[2] = 0;
        con c {s in {0,1,2}: x[s].sol > 0} : x[s] >= 1;
        MIN obj = x[0];
        expand;
        solve;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def drop(*args):
    """
    Drops the specified constraints or constraint groups from model

    Parameters
    ----------
    args : :class:`sasoptpy.Constraint`, :class:`sasoptpy.ConstraintGroup`
        Constraints to be dropped

    Examples
    --------

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x', lb=1)
    >>>     y = so.Variable(name='y', lb=0)
    >>>     c = so.Constraint(sm.sqrt(x) >= 5, name='c')
    >>>     o = so.Objective(x + y, sense=so.MIN, name='obj')
    >>>     s = solve()
    >>>     drop(c)
    >>>     o2 = so.Objective(x, sense=so.MIN, name='obj2')
    >>>     s2 = solve()
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x >= 1;
        var y >= 0;
        con c : sqrt(x) >= 5;
        min obj = x + y;
        solve;
        drop c;
        min obj2 = x;
        solve;
    quit;

    See also
    --------
    :func:`sasoptpy.actions.restore`

    """
    pass


@sasoptpy.containable(standalone=False)
def restore(*args):
    """
    Restores dropped constraint and constraint groups

    Parameters
    ----------
    args : :class:`sasoptpy.Constraint`, :class:`sasoptpy.ConstraintGroup`
        Constraints to be restored

    Examples
    --------

    >>> with so.Workspace('w') as w:
    >>>     x = so.Variable(name='x', lb=-1)
    >>>     set_objective(x**3, name='xcube', sense=so.minimize)
    >>>     c = so.Constraint(x >= 1, name='xbound')
    >>>     solve()
    >>>     drop(c)
    >>>     solve()
    >>>     restore(c)
    >>>     solve()
    >>> print(so.to_optmodel(w))
    proc optmodel;
        var x >= -1;
        MIN xcube = (x) ^ (3);
        con xbound : x >= 1;
        solve;
        drop xbound;
        solve;
        restore xbound;
        solve;
    quit;

    See also
    --------
    :func:`sasoptpy.actions.drop`

    """
    pass


@sasoptpy.containable(standalone=False)
def union(*args):
    """
    Aggregates the specified sets and set expressions

    Parameters
    ----------
    args : :class:`sasoptpy.abstract.Set` and :class:`sasoptpy.abstract.InlineSet`
        Objects to be aggregated

    Examples
    --------

    >>> from sasoptpy.actions import union, put_item
    >>> with so.Workspace('w') as w:
    >>>     n = so.Parameter(name='n', value=11)
    >>>     S = so.Set(name='S', value=so.exp_range(1, n))
    >>>     T = so.Set(name='T', value=so.exp_range(n+1, 20))
    >>>     U = so.Set(name='U', value=union(S, T))
    >>>     put_item(U, names=True)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        num n = 11;
        set S = 1..n;
        set T = n+1..20;
        set U = S union T;
        put U=;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def diff(left, right):
    """
    Gets the difference between set and set expressions

    Parameters
    ----------
    left : :class:`sasoptpy.abstract.Set`
        Left operand
    right : :class:`sasoptpy.abstract.Set`
        Right operand

    Examples
    --------

    >>> from sasoptpy.actions import diff, put_item
    >>> with so.Workspace('w') as w:
    >>>     S = so.Set(name='S', value=so.exp_range(1, 20))
    >>>     T = so.Set(name='T', value=so.exp_range(1, 15))
    >>>     U = so.Set(name='U', value=diff(S, T))
    >>>     put_item(U, names=True)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        set S = 1..19;
        set T = 1..14;
        set U = S diff T;
        put U=;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def substring(main_string, first_pos, last_pos):
    """
    Gets the substring of the specified positions

    Parameters
    ----------
    main_string : :class:`sasoptpy.abstract.Parameter` or string
        Main string
    first_pos : integer
        First position of the substring, starting from 1
    last_pos : integer
        Last position of the substring

    Examples
    --------

    >>> with so.Workspace('w') as w:
    >>>     p = so.Parameter(name='p', value='random_string', ptype=so.STR)
    >>>     r = so.Parameter(name='r', value=substring(p, 1, 6), ptype=so.STR)
    >>>     put_item(r)
    >>> print(so.to_optmodel(w))
    proc optmodel;
        str p = 'random_string';
        str r = substr(p, 1, 6);
        put r;
    quit;

    """
    pass


@sasoptpy.containable(standalone=False)
def use_problem(problem):
    """
    Changes the currently active problem

    Parameters
    ----------
    problem : :class:`sasoptpy.Model`
        Model to be activated

    Examples
    --------

    >>> from sasoptpy.actions import use_problem
    >>> with so.Workspace('w') as w:
    >>>     m = so.Model(name='m')
    >>>     m2 = so.Model(name='m2')
    >>>     use_problem(m)
    >>>     x = so.Variable(name='x')
    >>>     use_problem(m2)
    >>>     m.solve()
    >>>     m2.solve()
    >>> print(so.to_optmodel(w))
    proc optmodel;
        problem m;
        problem m2;
        use problem m;
        var x;
        use problem m2;
        use problem m;
        solve;
        use problem m2;
        solve;
    quit;
    """
    pass


condition = sasoptpy.structure.under_condition
inline_condition = sasoptpy.structure.inline_condition



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Test for symbolic conditions
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))
from util import assert_equal_wo_temps

class TestCondition(unittest.TestCase):
    """
    Unit tests for abstract conditions to be executed on the server
    """

    def test_regular_condition(self):

        from sasoptpy.util import iterate
        from sasoptpy.actions import condition

        x = so.VariableGroup(3, name='x')
        c = so.ConstraintGroup(None, name='c')
        with iterate([1, 2, 3], 's') as i:
            with condition(i != 2):
                c[i] = x[i] <= 3

        self.assertEqual(so.to_definition(c),
                         'con c {s in {1,2,3}: s NE 2} : x[s] <= 3;')

    def test_container_condition(self):

        from sasoptpy.util import iterate
        from sasoptpy.actions import condition, set_objective, solve, expand

        with so.Workspace(name='w') as w:

            x = so.VariableGroup(3, name='x')
            self.assertEqual(x[0].sym.get_conditions_str(), '')
            # solve
            x[0].set_value(1)
            x[1].set_value(5)
            x[2].set_value(0)

            c = so.ConstraintGroup(None, name='c')

            with iterate([0, 1, 2], 's') as i:
                with condition(x[i].sym > 0):
                    c[i] = x[i] >= 1

            set_objective(x[0], name='obj', sense=so.MIN)
            expand()
            solve()

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc(
            '''
            proc optmodel;
               var x {{0,1,2}};
               x[0] = 1;
               x[1] = 5;
               x[2] = 0;
               con c {s in {0,1,2}: x[s].sol > 0} : x[s] >= 1;
               MIN obj = x[0];
               expand;
               solve;
            quit;'''
        ))

    def test_nested_conditions(self):

        from sasoptpy.actions import condition

        S = so.Set(name='S')
        P = so.Set(name='P')
        x = so.VariableGroup(S, name='x')
        c = so.ConstraintGroup(None, name='c')
        d = so.ConstraintGroup(None, name='d')
        e = so.ConstraintGroup(None, name='e')
        for i in S:
            with condition(2 * i <= 5):
                t = i * x[i] <= 5
                c[i] = t
                with condition(i**2 >= 3):
                    d[i] = x[i] - i >= 5
                with condition((i >= 4) | (i <= 2)):
                    e[i] = x[i] >= 0

        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o9 in S: 2.0 * o9 <= 5} : o9 * x[o9] <= 5;')

        assert_equal_wo_temps(
            self, so.to_definition(d),
            'con d {o9 in S: 2.0 * o9 <= 5 and (o9) ^ (2) >= 3} : x[o9] - o9 >= 5;')

        assert_equal_wo_temps(
            self, so.to_definition(e),
            'con e {o9 in S: 2.0 * o9 <= 5 and ((o9 >= 4) or (o9 <= 2))} : x[o9] >= 0;')

    def test_conditional_all_combinations(self):

        from sasoptpy.actions import condition

        S = so.Set(name='S')
        x = so.VariableGroup(S, name='x')

        # EQ
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(x[i].sym == 0):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
            con c {o6 in S: x[o6].sol EQ 0} : x[o6] >= 1;'''))

        # LE
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(x[i].sym <= 0.5):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
            con c {o6 in S: x[o6].sol <= 0.5} : x[o6] >= 1;'''))

        # LT
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(x[i].sym < 0.5):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
            con c {o6 in S: x[o6].sol < 0.5} : x[o6] >= 1;'''))

        # GE
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(x[i].sym >= 0.5):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
            con c {o6 in S: x[o6].sol >= 0.5} : x[o6] >= 1;'''))

        # GT
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(x[i].sym > 0.5):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
            con c {o6 in S: x[o6].sol > 0.5} : x[o6] >= 1;'''))

        # NE
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition(x[i].sym != 1):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
            con c {o6 in S: x[o6].sol NE 1} : x[o6] >= 1;'''))

        # AND
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition((x[i].sym != 1) & (x[i].sym <= 2)):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
            con c {o6 in S: ((x[o6].sol NE 1) and (x[o6].sol <= 2))} : x[o6] >= 1;'''))

        # OR
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition((x[i].sym < 1) | (x[i].sym > 2)):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
            con c {o6 in S: ((x[o6].sol < 1) or (x[o6].sol > 2))} : x[o6] >= 1;'''))

        # OR
        c = so.ConstraintGroup(None, name='c')
        for i in S:
            with condition((x[i].sym < 1) | (x[i].sym > 2)):
                c[i] = x[i] >= 1
        assert_equal_wo_temps(self, so.to_definition(c), cleandoc('''
                    con c {o6 in S: ((x[o6].sol < 1) or (x[o6].sol > 2))} : x[o6] >= 1;'''))



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

''' Install the SAS Optimization Modeling for Python (sasoptpy) '''

from setuptools import setup

__version__ = None
with open('./sasoptpy/version.py') as f:
    exec(f.read())

long_desc = '''
sasoptpy: SAS Optimization Interface for Python

- Repository: https://github.com/sassoftware/sasoptpy
- Issues: https://github.com/sassoftware/sasoptpy/issues
- Releases: https://github.com/sassoftware/sasoptpy/releases
- Community: https://communities.sas.com/t5/Mathematical-Optimization/bd-p/operations_research

Copyright (C) SAS Institute Inc.
'''

setup(
    name='sasoptpy',
    version=__version__,
    packages=['sasoptpy'],
    description='sasoptpy: SAS Optimization Interface for Python',
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url='https://github.com/sassoftware/sasoptpy',
    author='Sertalp B. Cay (SAS Institute)',
    author_email='Sertalp.Cay@sas.com',
    license='Apache v2.0',
    install_requires=[
        'pandas',
        'swat',
        'saspy',
        'numpy'
        ],
    setup_requires=[
        'numpy'
        ],
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Mathematics"
    ],
    include_package_data=True,
)



import sasoptpy as so
import sasoptpy.abstract.math as sm


def test(cas_conn):

    m = so.Model(name='nlpse02', session=cas_conn)
    N = m.add_parameter(name='N', init=1000)
    x = m.add_variables(so.exp_range(1, N), name='x', init=1)
    m.set_objective(
        so.expr_sum(-4*x[i]+3 for i in so.exp_range(1, N-1)) +
        so.expr_sum((x[i]**2 + x[N]**2)**2 for i in so.exp_range(1, N-1)),
        name='f', sense=so.MIN)

    m.add_statement('print x;', after_solve=True)
    m.solve(options={'with': 'nlp'}, verbose=True)
    print(m.get_solution_summary())
    if m.get_session_type() == 'CAS':
        print(m.response['Print1.PrintTable'].head())

    # Model 2
    so.reset()
    m = so.Model(name='nlpse02_2', session=cas_conn)
    N = m.add_parameter(name='N', init=1000)
    x = m.add_variables(so.exp_range(1, N), name='x', lb=1, ub=2)
    m.set_objective(
        so.expr_sum(sm.cos(-0.5*x[i+1] - x[i]**2) for i in so.exp_range(
            1, N-1)), name='f2', sense=so.MIN)
    m.add_statement('print x;', after_solve=True)
    m.solve(verbose=True, options={'with': 'nlp', 'algorithm': 'activeset'})
    print(m.get_solution_summary())

    return m.get_objective_value()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for literal statements.
"""

import unittest
import sasoptpy as so
from inspect import cleandoc


class TestLiteral(unittest.TestCase):
    """
    Unit tests for literal statements
    """

    def test_literal_expr(self):
        literal = so.LiteralStatement('expand;')
        self.assertEqual(so.to_expression(literal), 'expand;')

    def test_use_problem(self):
        from sasoptpy.actions import use_problem
        with so.Workspace('w') as w:
            m = so.Model(name='m')
            m2 = so.Model(name='m2')
            use_problem(m)
            x = so.Variable(name='x')
            use_problem(m2)
            m.solve()
            m2.solve()

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               problem m;
               problem m2;
               use problem m;
               var x;
               use problem m2;
               use problem m;
               solve;
               use problem m2;
               solve;
            quit;'''))

    def test_union(self):
        from sasoptpy.actions import union, put_item
        with so.Workspace('w') as w:
            n = so.Parameter(name='n', value=11)
            S = so.Set(name='S', value=so.exp_range(1, n))
            T = so.Set(name='T', value=so.exp_range(n+1, 20))
            U = so.Set(name='U', value=union(S, T))
            put_item(U, names=True)

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               num n = 11;
               set S = 1..n;
               set T = n+1..20;
               set U = S union T;
               put U=;
            quit;'''))

    def test_diff(self):
        from sasoptpy.actions import diff, put_item
        with so.Workspace('w') as w:
            S = so.Set(name='S', value=so.exp_range(1, 20))
            T = so.Set(name='T', value=so.exp_range(1, 15))
            U = so.Set(name='U', value=diff(S, T))
            put_item(U, names=True)

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               set S = 1..19;
               set T = 1..14;
               set U = S diff T;
               put U=;
            quit;'''))

    def test_substring(self):
        from sasoptpy.actions import substring, put_item
        with so.Workspace('w') as w:
            p = so.Parameter(name='p', value='random_string', ptype=so.STR)
            r = so.Parameter(name='r', value=substring(p, 1, 6), ptype=so.STR)
            put_item(r)

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               str p = 'random_string';
               str r = substr(p, 1, 6);
               put r;
            quit;'''))
